% FLENG compiler - compiler core

main(['-v']) :-
    fleng_version(V),
    target_architecture(Arch),
    format('fl2~a ~a\n', [Arch, V]),
    halt(0).
main(['-I', Dir|Args]) :-
    asserta(include_path(Dir)),
    main(Args).
main(['-w'|Args]) :-
    assertz(no_warnings),
    main(Args).
main(['-d'|Args]) :-
    assertz(dump_analysis_info),
    main(Args).
main([Infile, Outfile]) :-
    op(200, fx, !),
    op(700, xfx, '\\=:='),
    op(500, yfx, '><'),
    op(400, yfx, '\\\\'),
    op(700, yfx, '@'),
    op(900, xfy, '&'),
    op(880, xfx, ':='),
    op(880, xfx, '<='),
    op(880, xfx, '+='),
    op(880, xfx, '-='),
    op(880, xfx, '*='),
    op(880, xfx, '/='),
    op(880, xfx, '<=='),
    op(880, xfx, '=>'),
    op(0, xfx, '^'),
    g_assign(atom_counter, 0),
    g_assign(literal_counter, 0),
    g_assign(symloc_counter, 0),
    assertz(include_path('.')),
    see(Infile),
    read_program(Prg),
    seen,
    collect_groups(Prg, 0, Groups),
    open(Outfile, write, Out),
    g_assign(output, Out),
    init_optimizer,
    generate_files(Infile, Groups, Out).
main(_) :- 
    target_architecture(Arch),
    format(user_error, 'usage: fl2~a [-w] [-d] [-v] INFILE OUTFILE\n', 
        [Arch]),
    halt(1).

generate_files(Infile, Groups, Out) :-
    init_asm(Infile, Out),
    analyze_calls(Groups),
    collect_inlining_candidates(Groups),
    emit_foreign_files(Out),
    compile_program(Groups, 0, _),
    compile_drop_sequences,
    finalize_optimizer,
    dump_atoms(0, Out),
    dump_data(Out),
    dump_literals(Out),
    asm(end, Out),
    close(Out),
    halt(0).

% utilities

read_program(PRG) :-
    read_term(TERM, [variable_names(VS), singletons(SL)]),
    check_max_env(VS, TERM),
    report_singletons(SL, TERM),
    numbervars(TERM),
    read_program(TERM, VS, PRG).

check_max_env(VS, _) :-
    length(VS, N),
    N =< 32.
check_max_env(_, T) :-
    error('too many variables in clause: ~q\n', [T]).

read_program(end_of_file, _, []).
read_program(-DECL, _, PRG) :-
    process_declaration(DECL, PRG).
read_program(TERM, VS, [t(TERM, VS, SVs)|PRG]) :-
    collect_singletons(TERM, SVs),
    read_program(PRG).

report_singletons([], _).
report_singletons(_, _) :- no_warnings.
report_singletons(SL, _) :-
    format(user_error, 'Warning: singleton variables ', []),
    member(Name = _, SL),
    format(user_error, '~a ', [Name]),
    fail.
report_singletons(_, Term) :-
    format(user_error, 'in term:\n\n  ~q\n', [Term]).

collect_singletons(T, Vs) :- 
    collect_vars(T, [], Vs1),
    remove_dups(Vs1, [], Vs).

collect_vars(A, Vs, Vs) :- atomic(A).
collect_vars('$VAR'(I), Vs, [I|Vs]).
collect_vars([A|B], Vs1, Vs) :-
    collect_vars(A, Vs1, Vs2),
    collect_vars(B, Vs2, Vs).
collect_vars(T, Vs1, Vs) :-
    callable(T),
    T =.. [_|Args],
    collect_vars(Args, Vs1, Vs).
collect_vars(T, _, _) :- error('invalid term: ~q\n', [T]).

remove_dups([], _, []).
remove_dups([I|L1], Drop, L) :- 
    member(I, L1),
    remove_dups(L1, [I|Drop], L).
remove_dups([I|L1], Drop, L) :-
    member(I, Drop),
    remove_dups(L1, Drop, L).
remove_dups([I|L1], Drop, [I|L]) :-
    remove_dups(L1, Drop, L).

process_declaration(initialization(NAME), Prg) :-
    assertz(module_init(NAME)),
    read_program(Prg).
process_declaration(synthesized(N1/A1, N2/A2), Prg) :-
    atom(N1), atom(N2),
    integer(A1), integer(A2),
    assertz(synthesized_name(N1/A1, N2/A2)),
    read_program(Prg).
process_declaration(include(Fname), Prg) :-
    atom(Fname),
    current_input(In),
    find_include_file(Fname, Fname2),
    see(Fname2),
    read_program(Included),
    seen,
    set_input(In),
    append(Included, Next, Prg),
    read_program(Next).
process_declaration(foreign(Name), PRG) :-
    assertz(foreign_file(Name)),
    read_program(PRG).
process_declaration(DECL, _) :-
    error('invalid declaration: ~q\n', [DECL]).

register_cmdline(Str) :- 
    atom(Str), 
    atom_codes(Str, C),
    assertz(cmdline_init(C)).
register_cmdline(Str) :- assertz(cmdline_init(Str)).

find_include_file(Fname1, Fname) :-
    include_path(Dir),
    format_to_atom(Fname, '~a/~a', [Dir, Fname1]),
    file_exists(Fname), 
    !.
find_include_file(Fname, _) :-
    error('included file does not exist: ~a\n', [Fname]).

error(Fstr, Args) :-
    format(user_error, Fstr, Args),
    halt(1).

collect_groups([], _, []).
collect_groups([T|PRG], I, [pdef(NA, I, EL, [T|G])|GROUPS]) :-
    T = t(TERM, VS, _),
    length(VS, EL),
    term_head(TERM, NA, _, _),
    filter_group(PRG, NA, G, NG),
    assertz(pdef(NA, I)),
    I2 is I + 1,
    collect_groups(NG, I2, GROUPS).

filter_group([], _, [], []).
filter_group([T|PRG], NA, [T|G], NG) :-
    T = t(TERM, _, _),
    term_head(TERM, NA, _, _),
    filter_group(PRG, NA, G, NG).
filter_group([T|PRG], NA, G, [T|NG]) :-
    filter_group(PRG, NA, G, NG).

term_head((H :- _), NA, EA, ARGS) :-
    !,
    term_head(H, NA, EA, ARGS).
term_head(H, N/A, EA, Args) :-
    H =.. [N|Args1],    
    compress_args(Args1, Args),
    length(Args1, A),
    length(Args, EA).

emit(Str, Out) :- write(Out, Str).
emit(Fstr, Args, Out) :- format(Out, Fstr, Args).

compress_args([A1, A2, A3, A4, A5|As], [A1, A2, A3, [A4, A5|As]]).
compress_args(As, As).

dbg(X) :- format('~q\n', [X]).
dbg(F, X) :- format(F, [X]).

label(Lc1, Lc, Lc1) :- Lc is Lc1 + 1.

register_data(Lc1, Lc, Lc1) :-
    Lc is Lc1 + 1,
    assertz(r_data(Lc1)).

dump_data(Out) :-
    r_data(Ld),
    asm(data(Ld), Out),
    fail.
dump_data(_).

dump_literals(Out) :-
    literal(X, L),
    dump_literal(X, L, Out),
    fail.
dump_literals(_).

% atoms

register_atom(A, I) :- r_atom(A, I), !.
register_atom(A, I) :-
    g_read(atom_counter, I),
    I2 is I + 1,
    g_assign(atom_counter, I2),
    assertz(r_atom(A, I)).

dump_atoms(I, Out) :-
    r_atom(A, I),
    asm(defatom(I, A), Out),
    I2 is I + 1,
    dump_atoms(I2, Out).
dump_atoms(_, _).

atom_count(I) :- g_read(atom_counter, I).

% compilation entry point

compile_program([], Lc, Lc).
compile_program([pdef(NA, _, _, _)|GROUPS], Lc1, Lc) :-
    (\+root_caller(NA, []); inline_candidate(NA, _, _)),
    !,
    retract(pdef(NA, _)),
    debug_compiler(prune, NA),
    compile_program(GROUPS, Lc1, Lc).
compile_program([pdef(NA, I, EL, G)|GROUPS], Lc1, Lc) :-
    compile_group(NA, I, EL, G, Lc1, Lc2),
    compile_program(GROUPS, Lc2, Lc).

% clause

compile_clause((H :- B), Ss, ForceArg1, Fail, Lc1, Lc) :-
    !,
    term_head(H, N/_, EA, Args),
    H2 =.. [N|Args],
    compile_head(H2, Ss, Fail, ForceArg1, Lc1, Lc2, Bound, Forced),
    count_bound_refs(Bound, AEnv1),
    analyze(B, Bound, _, AEnv1, AEnv),
    debug_compiler(type, (H :- B), AEnv),
    compile_body(B, EA, Ss, Fail, Bound, Forced, AEnv, Lc2, Lc).
compile_clause(C, _, _, _, _, _) :-
    error('invalid clause: ~q\n', [C]).

count_bound_refs([], []).
count_bound_refs([V|L], [refs(V, 1)|L2]) :- count_bound_refs(L, L2).

compile_head(H, Ss, Fail, ForceArg1, Lc1, Lc, Bound, Forced) :- 
    H =.. [_|Args],
    compile_matches(Args, 0, Ss, Fail, ForceArg1, Lc1, Lc, [], Bound, [], Forced).

% matching

compile_matches([], _, _, _, _, Lc, Lc, S, S, F, F).
compile_matches(['$VAR'(V)|Args], I, Ss, Fail, _, Lc1, Lc, S1, S, F1, F) :-
    member(V, Ss),
    I2 is I + 1,
    compile_matches(Args, I2, Ss, Fail, yes, Lc1, Lc, S1, S, F1, F).
compile_matches(['$VAR'(V)|Args], I, Ss, Fail, _, Lc1, Lc, S1, S, F1, F) :-
    \+member(V, Ss),
    \+member(V, S1),
    gen(putenv_arg(I, V)),
    I2 is I + 1,
    compile_matches(Args, I2, Ss, Fail, yes, Lc1, Lc, [V|S1], S, F1, F).
compile_matches([X|Args], I, Ss, Fail, FA1, Lc1, Lc, S1, S, F1, F) :-
    gen(getarg(I)),
    compile_match(X, Ss, Fail, FA1, Lc1, Lc2, S1, S2, F1, F2),
    I2 is I + 1,
    compile_matches(Args, I2, Ss, Fail, yes, Lc2, Lc, S2, S, F2, F).

compile_match('!'('$VAR'(VI)), _, _, _, _, _, Seen, _, _, _) :-
    member(VI, Seen),
    error('multiple references of variable in head: ~d\n', [VI]).
compile_match('!'('$VAR'(VI)), Ss, Fail, FA, Lc1, Lc, S, S, F, F) :-
    member(VI, Ss),
    drop_sequence(S, [], jump(Fail), Lc1, Lc, Ld),
    emit_force(FA, Ld).
compile_match('!'('$VAR'(VI)), _, Fail, FA, Lc1, Lc, S, [VI|S], F, [VI|F]) :-
    drop_sequence(S, [], jump(Fail), Lc1, Lc, Ld),
    emit_force(FA, Ld),
    gen(bind(VI)).
compile_match('!'(X), _, _, _, _, _, _, _, _, _) :-
    error('misplaced "!" operator: ~q\n', '?'(X)).
compile_match(N, _, Fail, FA, Lc1, Lc, S, S, F, F) :-
    integer(N),
    drop_sequence(S, [], jump(Fail), Lc1, Lc, Ld),
    emit_force(FA, Ld),
    gen(isint(N, Fail, Ld)).
compile_match([], _, Fail, FA, Lc1, Lc, S, S, F, F) :-
    % swipl doesn't treat "[]" as atom...
    register_atom('[]', AI),
    drop_sequence(S, [], jump(Fail), Lc1, Lc, Ld),
    emit_force(FA, Ld),
    gen(isatom(AI, Fail, Ld, '[]')).
compile_match(A, _, Fail, FA, Lc1, Lc, S, S, F, F) :-
    atom(A),
    register_atom(A, AI),
    drop_sequence(S, [], jump(Fail), Lc1, Lc, Ld),
    emit_force(FA, Ld),
    gen(isatom(AI, Fail, Ld, A)).
compile_match([X|Y], Ss, Fail, FA, Lc1, Lc, S1, S, F1, F) :-
    drop_sequence(S1, [], jump(Fail), Lc1, Lc2, Ld),
    emit_force(FA, Ld),
    gen(islist(Fail, Ld)),
    compile_match(X, Ss, Fail, yes, Lc2, Lc3, S1, S2, F1, F2),
    gen(pop),
    compile_match(Y, Ss, Fail, yes, Lc3, Lc, S2, S, F2, F).
compile_match('$VAR'(VI), _, _, _, _, _, Seen, _, _, _) :-
    member(VI, Seen),
    error('multiple references of variable in head: ~d\n', [VI]).
compile_match('$VAR'(VI), Ss, _, _, Lc, Lc, S, S, F, F) :-
    member(VI, Ss).
compile_match('$VAR'(VI), _, _, _, Lc, Lc, S, [VI|S], F, F) :-
    gen(bind(VI)).
compile_match('{}'(Xs), Ss, Fail, FA, Lc1, Lc, S1, S, F1, F) :-
    tuple_args(Xs, [Arg1|Args]),
    length(Args, N),
    drop_sequence(S1, [], jump(Fail), Lc1, Lc2, Ld),
    emit_force(FA, Ld),
    gen(istuple(N, Fail, Ld)),
    compile_match(Arg1, Ss, Fail, yes, Lc2, Lc3, S1, S2, F1, F2),
    gen(pop),
    compile_match(Args, Ss, Fail, yes, Lc3, Lc, S2, S, F2, F).
compile_match(T, Ss, Fail, FA, Lc1, Lc, S1, S, F1, F) :-
    callable(T),
    T =.. [Head|Args],
    length(Args, N),
    drop_sequence(S1, [], jump(Fail), Lc1, Lc2, Ld),
    emit_force(FA, Ld),
    gen(istuple(N, Fail, Ld)),
    compile_match(Head, Ss, Fail, yes, Lc2, Lc3, S1, S2, F1, F2),
    gen(pop),
    compile_match(Args, Ss, Fail, yes, Lc3, Lc, S2, S, F2, F).
compile_match(Arg, _, _, _, _, _, _, _, _, _) :-
    error('invalid match: ~q\n', [Arg]).

emit_force(no, _) :- gen(deref).
emit_force(yes, Fail) :- gen(force(Fail)).

synthesized(NA1, NA2) :- 
    synthesized_name(NA1, NA3), 
    !,
    synthesized(NA3, NA2).
synthesized(NA1, NA1).

tuple_args(T, L) :-
    tuple_args2(T, L),
    length(L, N),
    N =< 7.
tuple_args(T, _) :-
    error('tuple exceeds maximum size: ~q\n', [T]).

tuple_args2((X, Y), [X|Args]) :- !, tuple_args2(Y, Args).
tuple_args2(X, [X]).
    
% bodies

compile_body(B, Arity, Ss, Fail, Bound, Forced, AEnv, Lc1, Lc) :-
    reduce_inline_calls(B, B2, Ss, Ss2),
    extract_computations(B2, Bc, [], L),
    merge_computations(L, AEnv, L2),
    reconstruct_computations(L2),
    debug_compiler(expr, L),
    compile_primitives(Bc, AEnv, Ss2, Fail, Forced, F2, Bound, B3, Lc1, Lc2, [], User),
    reverse(User, User1),
    reorder_goals(User1, User2),
    compile_user_goals(User2, Arity, B3, AEnv, Ss2, F2, Lc2, Lc).

reorder_goals([], []).
reorder_goals([G], [G]).
reorder_goals([G0|Gs], GR) :-
    append(Gs, [G0], GR).

reorder_goals([G], [], G).
reorder_goals([G1|Gs], [G1|GR], G0) :-
    reorder_goals(Gs, GR, G0).

compile_primitives((X, Y), AEnv, Ss, Fl, Fd1, Fd, Env1, Env, Lc1, Lc, U1, U) :-
    !,
    compile_primitives(X, AEnv, Ss, Fl, Fd1, Fd2, Env1, Env2, Lc1, Lc2, U1, U2),
    compile_primitives(Y, AEnv, Ss, Fl, Fd2, Fd, Env2, Env, Lc2, Lc, U2, U).
compile_primitives(true, _, _, _, F, F, E, E, Lc, Lc, U, U).
compile_primitives(X = Y, AEnv, Ss, Fl, F1, F, E1, E, Lc1, Lc, U1, U) :-
    compile_primitives('$unify'(X, Y, '$VAR'('_')), AEnv, Ss, Fl, F1, F, E1, E, 
        Lc1, Lc, U1, U).
compile_primitives(unify(X, Y, Z), AEnv, Ss, Fl, F1, F, E1, E, Lc1, Lc, U1, U) :-
    compile_primitives('$unify'(X, Y, Z), AEnv, Ss, Fl, F1, F, E1, E, 
        Lc1, Lc, U1, U).
compile_primitives('$suspend', AEnv, _, _, F, F, E, E, Lc1, Lc, U, U) :-
    drop_sequence(E, AEnv, suspend, Lc1, Lc, Ld),
    gen(suspending(Ld)).
compile_primitives(label(NA, R), AEnv, Ss, _, Fd1, Fd, E1, E, Lc, Lc, U, U) :-
    resolve_label(NA, L),
    gen(code_label(NA, L)),
    compile_compute_result(R, AEnv, Ss, Fd1, Fd, E1, E).
compile_primitives('$guard'('$VAR'(V), true), AEnv, _, Fail, F, F, E, E, Lc1, 
    Lc, U, U) :- 
    member(temp(V), AEnv),
    member(type(V, a), AEnv),
    drop_sequence(E, AEnv, jump(Fail), Lc1, Lc, Ld),
    gen([getenv(V), if_true(Ld)]).
compile_primitives('$guard'('$VAR'(V), false), AEnv, _, Fail, F, F, E, E, Lc1, 
    Lc, U, U) :- 
    member(temp(V), AEnv),
    member(type(V, a), AEnv),
    drop_sequence(E, AEnv, jump(Fail), Lc1, Lc, Ld),
    gen([getenv(V), if_false(Ld)]).
compile_primitives('$guard'(X, Y), AEnv, Ss, Fail, F, F, E1, E, Lc1, Lc, U, U) :-
    drop_sequence(E1, AEnv, jump(Fail), Lc1, Lc, Ld),
    compile_arguments([X, Y], any, putparam, Ss, F, AEnv, E1, E, Build),
    gen(guard(Ld)),
    unref_arguments(Build).
compile_primitives('$idle'(R), AEnv, Ss, _, Fd1, Fd, E1, E, Lc1, Lc, U, U) :-
    register_data(Lc1, Lc, Ld),
    gen(idle(Ld)),
    compile_compute_result(R, AEnv, Ss, Fd1, Fd, E1, E).    
compile_primitives(foreign_call(T), AEnv, Ss, _, F, F, E1, E, Lc, Lc, U, U) :-
    compile_foreign_call(T, AEnv, Ss, F, E1, E).
compile_primitives(T, _, _, _, F, F, E, E, Lc, Lc, U, U) :-
    callable(T),
    T =.. ['$asm'|Args],
    extract_asm_params(Args, E, Asm),
    gen([asm_enter|Asm]).
compile_primitives(compute(Op, A, B, R), _, _, _, _, _, _, _, _, _, _, _) :-
    R \= '$VAR'(_),
    error('computation result must be a variable: ~q\n', [compute(Op, A, B, R)]).
compile_primitives(compute(Op, A, R),_, _, _, _, _, _, _, _, _, _, _) :-
    R \= '$VAR'(_),
    error('computation result must be a variable: ~q\n', [compute(Op, A, R)]).
compile_primitives(compute(Op, A, B, R), AEnv, Ss, _, Fd1, Fd, E1, E, Lc, 
        Lc, U, U) :-
    memberchk(Op, [=, >, <]),
    !,
    check_computation(Op, Types),
    length(Types, 3),
    compile_arguments([A, B], Types, putparam, Ss, Fd1, AEnv, E1, E2, Build),
    gen(compute(Op)),
    unref_arguments(Build),
    compile_compute_result(R, AEnv, Ss, Fd1, Fd, E2, E).
compile_primitives(compute(Op, A, R), AEnv, Ss, _, Fd1, Fd, E1, E, Lc, Lc, U, U) :-
    atom(Op),
    check_computation(Op, Types),
    length(Types, 2),
    compile_compute_arguments([A], Types, Ss, Fd1, AEnv, E1, E2),
    gen([compute(Op), to_int]),
    compile_compute_result(R, AEnv, Ss, Fd1, Fd, E2, E).
compile_primitives(compute(Op, A, B, R), AEnv, Ss, _, Fd1, Fd, E1, E, Lc, 
        Lc, U, U) :-
    atom(Op),
    check_computation(Op, Types),
    length(Types, 3),
    compile_compute_arguments([A, B], Types, Ss, Fd1, AEnv, E1, E2),
    gen([compute(Op), to_int]),
    compile_compute_result(R, AEnv, Ss, Fd1, Fd, E2, E).
compile_primitives(compute(Op, A, R), AEnv, Ss, _, Fd1, Fd, E1, E, Lc, 
        Lc, U, U) :-
    atom(Op),
    check_computation(Op, Types),
    length(Types, 2),
    compile_compute_arguments([A], Types, Ss, Fd1, AEnv, E1, E2),
    gen([compute(Op), to_int]),
    compile_compute_result(R, AEnv, Ss, Fd1, Fd, E2, E).
compile_primitives('$expr'(Exp), AEnv, Ss, _, Fd1, Fd, E1, E, Lc, Lc, U, U) :-
    compile_compute_expression(Exp, AEnv, Ss, Fd1, Fd, E1, E).
compile_primitives('$address_of'(L, R), AEnv, Ss, _, Fd1, Fd, E1, E, Lc, Lc, U, U) :-
    check_address(L),
    gen([addr_of(L), to_int]),
    compile_compute_result(R, AEnv, Ss, Fd1, Fd, E1, E).
compile_primitives(T, AEnv, Ss, _, Fd1, Fd, Env1, Env, Lc, Lc, U, U) :-
    T =.. [N|Args1],
    length(Args1, A),
    primitive(N/A, Op, result, Types),
    split_args(Args1, Args, R),
    compile_arguments(Args, Types, putparam, Ss, Fd1, AEnv, Env1, Env2, Build),
    gen(Op),
    unref_arguments(Build),
    compile_compute_result(R, AEnv, Ss, Fd1, Fd, Env2, Env).
compile_primitives(T, AEnv, Ss, _, Fd, Fd, Env1, Env, Lc, Lc, U, U) :-
    T =.. [N|Args],
    length(Args, A),
    primitive(N/A, Op, _, Types),
    compile_arguments(Args, Types, putparam, Ss, Fd, AEnv, Env1, Env, Build),
    gen(Op),
    unref_arguments(Build).
compile_primitives(U, _, _, _, Fd, Fd, Env, Env, Lc, Lc, User, [U|User]).

check_address(L) :- atom(L).
check_address(L) :- error('invalid address: ~q\n', [L]).

resolve_label(N/A, I) :- pdef(N/A, I).
resolve_label(NA, _) :-
    error('attempt to take address of undefined user goal: ~w\n', [NA]).

compile_foreign_call(T, AEnv, Ss, F, E1, E) :-
    callable(T),
    T =.. [Name|Args],
    length(Args, N),
    N < 5,
    compile_arguments(Args, any, putparam, Ss, F, AEnv, E1, E, Build),
    gen(foreign_call(Name, N)),
    unref_arguments(Build).
compile_foreign_call(T, _, _, _, _, _, _, _, _, _) :-
    error('invalid foreign call: ~q\n', [T]).

compile_user_goals([], Arity, Env, AEnv, _, _, Lc1, Lc) :-
    drop_sequence(Env, AEnv, done(Arity), Lc1, Lc, Ld),
    gen(jump(Ld)).
compile_user_goals(['$invoke'(PDef, Args)], A, Env, AEnv, Ss, Fd, Lc1, Lc) :-
    compile_drop_args(A),
    compile_value(PDef, p, Ss, Fd, AEnv, Env, Env2),
    gen(invoke_goal),
    compile_value(Args, x, Ss, Fd, AEnv, Env2, Env),
    compile_invoke_arguments(Lc1, Lc2),
    drop_sequence(Env2, AEnv, invoke, Lc2, Lc, Ld),
    gen(jump(Ld)).
compile_user_goals([G], Arity, Env, AEnv, Ss, _, Lc1, Lc) :-
    user_goal(G, N/A, I, Args),
    !,
    gen(tailgoal(I, N, A)),
    compile_drop_args(Arity),
    compile_arguments(Args, any, putarg, Ss, noderef, AEnv, Env, Env2, _),
    drop_sequence(Env2, AEnv, jump, Lc1, Lc, Ld),
    gen(jump(Ld)).
compile_user_goals([G|Gs], Arity, Env, AEnv, Ss, Fd, Lc1, Lc) :-
    user_goal(G, N/A, I, Args),
    gen(open_goal(I, N, A)),
    compile_arguments(Args, any, putarg, Ss, noderef, AEnv, Env, Env2, _),
    compile_user_goals(Gs, Arity, Env2, AEnv, Ss, Fd, Lc1, Lc).

user_goal(G, NA, I, Args) :-
    called_user_goal(G, NA, Args),
    pdef(NA, I).
user_goal(G, _, _, _) :-
    error('call to undefined goal: ~q\n', [G]).

called_user_goal(G, N/A, Args) :-
    callable(G),
    G =.. [N|Args1],
    compress_args(Args1, Args),
    length(Args1, A).

compile_value(T, Type, Ss, AEnv, Env1, Env) :-
    compile_value(T, Type, Ss, noderef, AEnv, Env1, Env).

compile_value(N, T, _, _, _, Env, Env) :-
    integer(N),
    (N < -16384; N > 16383),
    !,
    register_literal(N, Lit),
    gen(literal(Lit)),
    compile_check(T, i).
compile_value(N, T, _, _, _, Env, Env) :-
    integer(N),
    gen(mkint(N)),
    compile_check(T, i).
compile_value([], T, _, _, _, Env, Env) :-
    % SWI-prolog doesn't treat "[]" as an atom
    register_atom('[]', I),
    gen(atom(I, '[]')),
    compile_check(T, a).
compile_value(S, T, _, _, _, Env, Env) :-
    atom(S),
    register_atom(S, I),
    gen(atom(I, S)),
    compile_check(T, a).
compile_value(X, _, _, _, _, Env, Env) :-
    ground_value(X),
    !,
    register_literal(X, Lit),
    gen(literal(Lit)).
compile_value('$VAR'('_'), T, _, _, _, Env, Env) :-
    !,
    gen(mkvar),
    compile_check(T, v).
compile_value('$VAR'(I), T, Ss, _, _, Env, Env) :-
    memberchk(I, Ss),
    !,
    gen(mkvar),
    compile_check(T, v).
compile_value('$VAR'(I), T, _, Fd, AEnv, Env, Env) :-
    memberchk(I, Env),
    !,
    gen(getenv(I)),
    compile_deref(I, Fd),
    compile_check_var(T, I, AEnv).
compile_value('$VAR'(I), T, _, _, _, Env, [I|Env]) :-
    gen(bindvar(I)),
    compile_check(T, v).
compile_value([X|Y], T, Ss, Fd, AEnv, Env1, Env) :-
    compile_value(X, x, Ss, Fd, AEnv, Env1, Env2),
    gen(push),
    compile_value(Y, x, Ss, Fd, AEnv, Env2, Env),
    gen(mklist),
    compile_check(T, l).
compile_value('{}'(X), T, Ss, Fd, _, Env1, Env) :-
    tuple_args(X, Args),
    compile_value(Args, x, Ss, Fd, [], Env1, Env),
    length(Args, Arity),
    A2 is Arity - 1,
    gen(mktuple(A2)),
    compile_check(T, t).
compile_value(X, T, Ss, Fd, _, Env1, Env) :-
    callable(X),
    X =.. [Name|Args],
    register_atom(Name, I),
    compile_value(Args, x, Ss, Fd, [], Env1, Env),
    length(Args, Arity),
    gen(mktuple(I, Arity, Name)),
    compile_check(T, t).
compile_value(X, _, _, _, _, _, _) :-
    error('invalid value: ~q\n', [X]).

ground_value(X) :- 
    float(X),
    error('floating point literals are not supported: ~q\n', [X]).
ground_value(X) :- atomic(X), !.
ground_value([X|Y]) :- !, ground_value(X), ground_value(Y).
ground_value('{}'(X)) :- !, ground_value(X).
ground_value('$VAR'(_)) :- !, fail.
ground_value(T) :- 
    T =.. [_|Lst],
    ground_value(Lst).

compile_deref(_, noderef).
compile_deref(I, Fd) :- memberchk(I, Fd).
compile_deref(_, _) :- gen(deref).

compile_check(T, T) :- !.
compile_check(c, _) :- gen(check_cell).
compile_check(i, _) :- gen(check_integer).
compile_check(a, _) :- gen(check_atom).
compile_check(t, _) :- gen(check_tuple).
compile_check(l, _) :- gen(check_list).
compile_check(_, _).

compile_check_var(T, V, AEnv) :-
    memberchk(type(V, T2), AEnv),
    compile_check(T, T2).
compile_check_var(T, _, _) :- compile_check(T, x).

compile_arguments(Args, Types, Op, Ss, Fd, AEnv, Env1, Env, Build) :-
    compile_arguments(Args, Types, 0, Op, Ss, Fd, AEnv, Env1, Env, [], Build).

compile_arguments([], _, _, _, _, _, _, Env, Env, B, B).
compile_arguments(['$VAR'(V)|Args], any, I, putarg, Ss, Fd, AEnv, Env1, Env, B1, B) :-
    V \= '_',
    \+memberchk(V, Ss),
    memberchk(V, Env1),
    !,
    gen(putarg_env(I, V)),
    I2 is I + 1,
    compile_arguments(Args, any, I2, putarg, Ss, Fd, AEnv, Env1, Env, B1, B).
compile_arguments(['$VAR'(V)|Args], any, I, putarg, Ss, Fd, AEnv, Env1, Env, B1, B) :-
    V \= '_',
    \+memberchk(V, Ss),
    !,
    gen(putarg_bind(I, V)),
    I2 is I + 1,
    compile_arguments(Args, any, I2, putarg, Ss, Fd, AEnv, [V|Env1], Env, B1, B).
compile_arguments(['$VAR'(V)|Args], [T|Types], I, putparam, Ss, Fd, AEnv, Env1, Env, 
    B1, B) :-
    V \= '_',
    \+memberchk(V, Ss),
    memberchk(V, Env1),
    (T = x; T = v),
    !,
    gen(putparam_env(I, V)),
    I2 is I + 1,
    compile_arguments(Args, Types, I2, putparam, Ss, Fd, AEnv, Env1, Env, B1, B).
compile_arguments(['$VAR'(V)|Args], [T|Types], I, putparam, Ss, Fd, AEnv, Env1, Env, 
    B1, B) :-
    V \= '_',
    \+memberchk(V, Ss),
    \+memberchk(V, Env1),
    (T = x; T = v),
    !,
    gen(putparam_bind(I, V)),
    I2 is I + 1,
    compile_arguments(Args, Types, I2, putparam, Ss, Fd, AEnv, [V|Env1], Env, B1, B).
compile_arguments([X|Args], any, I, Op, Ss, Fd, AEnv, Env1, Env, B1, B) :-
    compile_value(X, x, Ss, Fd, AEnv, Env1, Env2),
    Instr =.. [Op, I],
    gen(Instr),
    I2 is I + 1,
    build_param(X, Op, I, Ss, B1, B2),
    compile_arguments(Args, any, I2, Op, Ss, Fd, AEnv, Env2, Env, B2, B).
compile_arguments([X|Args], [T|Types], I, Op, Ss, Fd, AEnv, Env1, Env, B1, B) :-
    compile_value(X, T, Ss, Fd, AEnv, Env1, Env2),
    Instr =.. [Op, I],
    gen(Instr),
    I2 is I + 1,
    build_param(X, Op, I, Ss, B1, B2),
    compile_arguments(Args, Types, I2, Op, Ss, Fd, AEnv, Env2, Env, B2, B).

% remember arguments to primitives that are created on the fly and need to be
% unref'd after use:
build_param('$VAR'('_'), putparam, I, _, B, [I|B]) :- 
    gen(addref).
build_param('$VAR'(V), putparam, I, Ss, B, [I|B]) :- 
    member(V, Ss),
    gen(addref).
build_param('$VAR'(_), putparam, _, _, B, B).
build_param('$VAR'(_), putarg, _, _, B, B) :-
    gen(addrefx).
build_param(X, putparam, I, _, B, [I|B]) :- 
    compound(X),
    \+ ground_value(X),
    gen(addref).
build_param(X, putarg, _, _, B, B) :-
    \+ ground_value(X),
    gen(addref).
build_param(_, _, _, _, B, B).

compile_invoke_arguments(Lc1, Lc) :-
    label(Lc1, Lc, Lf),
    compile_invoke_arguments2(0, Lf),
    gen([islist_t(Lf), puttailarg, label(Lf)]).
 
compile_invoke_arguments2(3, _).
compile_invoke_arguments2(I, Lf) :-
    gen([islist_n(Lf), putarg(I), addrefx, pop, deref]),
    I2 is I + 1,
    compile_invoke_arguments2(I2, Lf).

% computation optimization

check_computation(Op, Types) :- computation(Op, Types).
check_computation(Op, _) :- error('invalid computation: ~w\n', [Op]).

compile_compute_result('$VAR'('_'), _, _, Fd, Fd, Env, Env).
compile_compute_result('$VAR'(R), AEnv, _, Fd, [R|Fd], Env, [R|Env]) :-
    memberchk(temp(R), AEnv),
    gen(putenvu(R)).
compile_compute_result('$VAR'(R), _, Ss, Fd, Fd, Env, Env) :-
    memberchk(R, Ss).
compile_compute_result(R, _, Ss, Fd, Fd, Env1, Env) :-
    gen(putparam(0)),
    compile_value(R, x, Ss, [], Env1, Env),
    gen(assign).

compile_compute_expression(':='(V, X), AEnv, Ss, Fd1, Fd, E1, E) :-
    compile_compute_expression(X, AEnv, Ss, Fd1, Fd2, E1, E2),
    gen([push, to_int]),
    compile_compute_result(V, AEnv, Ss, Fd2, Fd, E2, E),
    gen(pop).
compile_compute_expression(X, AEnv, Ss, Fd1, Fd, E1, E) :-
    X \= '$VAR'(_),
    X =.. [Op|Args],
    compile_expr_arguments(Args, Ss, Fd1, Fd, AEnv, E1, E),
    gen(compute(Op)).
compile_compute_expression(N, _, _, Fd, Fd, E, E) :-
    integer(N),
    N >= -32768, N =< 32767, 
    gen(int_lit(N)).
compile_compute_expression(N, _, _, _, _, _, _) :-
    integer(N),
    error('integer literal out of range: ~d\n', [N]).
compile_compute_expression(X, AEnv, Ss, Fd, Fd, E1, E) :-
    compile_value(X, i, Ss, Fd, AEnv, E1, E),
    gen(intval).

compile_expr_arguments([X], Ss, Fd1, Fd, AEnv, E1, E2) :-
    compile_compute_expression(X, AEnv, Ss, Fd1, Fd, E1, E2).
compile_expr_arguments([X|L], Ss, Fd1, Fd, AEnv, E1, E) :-
    L \= [], 
    compile_compute_expression(X, AEnv, Ss, Fd1, Fd2, E1, E2),
    gen(push),
    compile_expr_arguments(L, Ss, Fd2, Fd, AEnv, E2, E).

compile_compute_arguments([X], [T|_], Ss, Fd, AEnv, Env1, Env) :-
    compile_value(X, T, Ss, Fd, AEnv, Env1, Env),
    gen(intval).
compile_compute_arguments([X|Args], [T|Types], Ss, Fd, AEnv, Env1, Env) :-
    compile_value(X, T, Ss, Fd, AEnv, Env1, Env2),
    gen([intval, push]),
    compile_compute_arguments(Args, Types, Ss, Fd, AEnv, Env2, Env).

extract_computations((X, Y), (X2, Y2), L1, L) :-
    extract_computations(X, X2, L1, L2),
    extract_computations(Y, Y2, L2, L).
extract_computations(X, C, L, [X-C|L]) :-
    functor(X, compute, A),
    A >= 2, A =< 4,
    X =.. [_, Op|_],
    computation(Op, _),
    \+memberchk(Op, [=, >, <]).
extract_computations(X, X, L, L).

merge_computations(CL1, AEnv, CL) :-
    select(C-V, CL1, CL2),
    computation_to_expr(C, Cb, R),
    keep_expr_ref(Cb, R, AEnv, Cc),
    select(C2-V2, CL2, CL3),
    replace_computation_argument(C2, R, Cc, C2b),
    !,
    V = true,
    merge_computations([C2b-V2|CL3], AEnv, CL).
merge_computations(CL, _, CL).

keep_expr_ref(X, V, AEnv, ':='('$VAR'(V), X)) :-
    memberchk(refs(V, N), AEnv),
    N > 2.
keep_expr_ref(X, _, _, X).

computation_to_expr(compute(Op, X, Y, '$VAR'(R)), T, R) :- T =.. [Op, X, Y].
computation_to_expr(compute(Op, X, '$VAR'(R)), T, R) :- T =.. [Op, X].
computation_to_expr('$expr'(':='('$VAR'(R), X)), X, R).

replace_computation_argument(compute(Op, '$VAR'(V), B, R), V, X, T) :-
    Exp =.. [Op, X, B],
    T = '$expr'(':='(R, Exp)).
replace_computation_argument(compute(Op, A, '$VAR'(V), R), V, X, T) :-
    Exp =.. [Op, A, X],
    T = '$expr'(':='(R, Exp)).
replace_computation_argument(compute(Op, '$VAR'(V), R), V, X, T) :-
    Exp =.. [Op, X],
    T = '$expr'(':='(R, Exp)).
replace_computation_argument('$expr'(Exp1), V, X, '$expr'(Exp2)) :-
    replace_computation_argument2(Exp1, V, X, Exp2).

replace_computation_argument2(X, _, _, X) :- atomic(X).
replace_computation_argument2([A|L], V, X, [A2|L2]) :-
    replace_computation_argument2(A, V, X, A2),
    replace_computation_argument2(L, V, X, L2).
replace_computation_argument2('$VAR'(V), V, X, X).
replace_computation_argument2('$VAR'(V2), V, _, '$VAR'(V2)) :- V \= V2.
replace_computation_argument2(T, V, X, T2) :-
    callable(T),
    T =.. [Op|Args],
    replace_computation_argument2(Args, V, X, Args2),
    T2 =.. [Op|Args2].

computation_result(compute(_, _, _, R), R).
computation_result(compute(_, _, R), R).
computation_result('$expr'(':='(R, _)), R).

reconstruct_computations([]).
reconstruct_computations([X-V|L]) :-
    V = X,
    reconstruct_computations(L).

% drop sequences

drop_sequence(Env, Type, Lc1, Lc, Ld) :-
    drop_sequence(Env, [], Type, Lc1, Lc, Ld).

drop_sequence(Env, AEnv, Type, Lc1, Lc, Ld) :-
    findall(V, member(temp(V), AEnv), MEnv),
    subtract(Env, MEnv, Env1),
    sort(Env1, Env2),
    reverse(Env2, SEnv),
    drop_sequence2(SEnv, Type, Lc1, Lc, Ld).

drop_sequence2([], jump(L), Lc, Lc, L) :- !.
drop_sequence2(SEnv, Type, Lc, Lc, Ld) :-
    drop_sequence(SEnv, Type, Ld), 
    !.
drop_sequence2(SEnv, Type, Lc1, Lc, Ld) :-
    label(Lc1, Lc, Ld),
    assertz(drop_sequence(SEnv, Type, Ld)).

compile_drop_sequences :-
    drop_sequence(Env, Type, Ld),
    compile_drop_sequence(Env, Type, Ld),
    fail.
compile_drop_sequences.

compile_drop_sequence([V|Vars], Type, Ld) :-
    gen([label(Ld), drop(V)]),
    drop_sequence(Vars, Type, Ld2),
    gen(jump(Ld2)),
    !.
compile_drop_sequence([_|Vars], _, _) :-
    member(V, Vars),
    gen(drop(V)),
    fail.
compile_drop_sequence([], _, Ld) :-
    gen(label(Ld)),
    fail.
compile_drop_sequence(_, done(A), _) :-
    compile_drop_args(A),
    gen(dropgoal),
    !.
compile_drop_sequence(_, yield, _) :-
    gen(yield),
    !.
compile_drop_sequence(_, jump, _) :-
    gen(tailjump),
    !.
compile_drop_sequence(_, invoke, _) :-
    gen(invoke_jump),
    !.
compile_drop_sequence(_, suspend, _) :-
    gen(suspend),
    !.
compile_drop_sequence(_, jump(L), _) :-
    gen(jump(L)),
    !.
compile_drop_sequence(_, Type, _) :-
    error('bad drop sequence type: ~q\n', [Type]).

compile_drop_args(A) :- compile_drop_args(A, 0).

compile_drop_args(I, I).
compile_drop_args(_, 4).
compile_drop_args(I, J) :-
    I > J,
    gen(droparg(J)),
    J2 is J + 1,
    compile_drop_args(I, J2).

% literals

register_literal(X, Lit) :- literal(X, Lit), !.
register_literal(N, Lit) :-
    integer(N),
    !,
    add_literal(N, Lit).
register_literal([X|Y], Lit) :-
    !,
    sub_literal(X, XL),
    sub_literal(Y, YL),
    add_literal([XL|YL], Lit).
register_literal('{}'((X, Y)), Lit) :-    
    !,
    tuple_args((X, Y), [Arg1|Args]),
    length(Args, A),
    register_literal(Args, AL),
    sub_literal(Arg1, A1L),
    maybe_add_literal(tuple(A, A1L, AL), Lit).
register_literal('{}'(X), Lit) :-
    % singleton tuple
    !,
    sub_literal(X, A1L),
    register_atom('[]', _), % in case [] is not yet registered
    maybe_add_literal(tuple0(A1L), Lit).
register_literal(X, Lit) :-
    X =.. [N|Args],
    length(Args, A),
    register_atom(N, NI),
    register_literal(Args, AL),
    maybe_add_literal(functor(NI, A, AL), Lit).
register_literal(X, _) :- error('bad literal: ~q\n', [X]).

maybe_add_literal(L, Lit) :- literal(L, Lit), !.
maybe_add_literal(L, Lit) :- add_literal(L, Lit).

sub_literal([], '[]') :- !, register_atom('[]', _).  % for swipl
sub_literal(X, X) :- atom(X), !, register_atom(X, _).
sub_literal(X, X) :- 
    integer(X),
    X >= -16384, X =< 16383,
    !.
sub_literal(X, label(Lit)) :- 
    integer(X),
    register_literal(X, Lit).
sub_literal(X, label(Lit)) :- register_literal(X, Lit).

add_literal(X, Lit) :-
    g_read(literal_counter, Lit),
    Lit2 is Lit + 1,
    g_assign(literal_counter, Lit2),
    assertz(literal(X, Lit)).

% helper operations

string_length([], 2) :- !.  % SWI-Prolog sillyness
string_length(S, Len) :- atom_length(S, Len).

unref_arguments(Build) :-
    member(P, Build),
    gen(unref(P)),
    fail.
unref_arguments(_).

register_symbol_loc(A, SL) :-
    g_read(symloc_counter, SL),
    SL2 is SL + 1,
    g_assign(symloc_counter, SL2),
    assertz(symbol_loc(SL, A)).

extract_asm_params([], _, [asm_exit]).
extract_asm_params([X|Args], Env, [asm(X)|Asm]) :-
    atom(X),
    extract_asm_params(Args, Env, Asm).
extract_asm_params([X|Args], Env, Asm) :-
    callable(X),
    X =.. [Code|Vars],
    deref_asm_vars(Vars, X, Env, 0, Asm, [asm(Code)|Asm2]),
    assign_asm_vars(Vars, X, Env, Asm2, Asm3),
    extract_asm_params(Args, Env, Asm3).
extract_asm_params([X|_], _, _) :-
    error('invalid argument to $asm primitive: ~w\n', [X]).

deref_asm_vars([], _, _, _, Asm, Asm).
deref_asm_vars(_, X, _, PR, _, _) :-
    PR > 3,
    error('too many variable references in $asm primitive: ~w\n', [X]).
deref_asm_vars(['$VAR'(_) = _|Vars], X, Env, PR, Asm1, Asm) :-
    deref_asm_vars(Vars, X, Env, PR, Asm1, Asm).
deref_asm_vars(['$VAR'(I)|Vars], X, Env, PR, Asm1, Asm) :-
    memberchk(I, Env),
    Asm1 = [x_putasmparam(I, PR)|Asm2],
    PR2 is PR + 1,
    deref_asm_vars(Vars, X, Env, PR2, Asm2, Asm).
deref_asm_vars(_, X, _, _, _, _) :-
    error('reference to undefined source variable in $asm primitive: ~w\n', [X]).

assign_asm_vars([], _, _, Asm, Asm).
assign_asm_vars(['$VAR'(I) = Reg|Vars], X, Env, Asm1, Asm) :-
    memberchk(I, Env),
    Asm1 = [x_getregister(Reg), x_assign(I)|Asm2],
    assign_asm_vars(Vars, X, Env, Asm2, Asm).
assign_asm_vars(['$VAR'(_)|Vars], X, Env, Asm1, Asm) :-
    assign_asm_vars(Vars, X, Env, Asm1, Asm).
assign_asm_vars(_, X, _, _, _) :-
    error('reference to undefined destination variable in $asm primitive: ~w\n', [X]).

debug_compiler(prune, N/A) :-
    dump_analysis_info,
    !,
    format(user_error, 'dropped: ~a/~d\n', [N, A]).
debug_compiler(expr, Term) :-
    Term \= [],
    dump_analysis_info,
    !,
    format(user_error, 'expr: ~w\n', [Term]).
debug_compiler(_, _).

debug_compiler(type, Term, AEnv) :-
    dump_analysis_info,
    !,
    format(user_error, 'types: ~q.\ntypes:  -> ~q\n', [Term, AEnv]).
debug_compiler(inline, Old, New) :-
    dump_analysis_info,
    !,
    format(user_error, 'inline: ~q\ninline:  -> ~q\n', [Old, New]).
debug_compiler(_, _, _).

debug_compiler(opt, Id, From, To) :-
    dump_analysis_info,
    !,
    format(user_error, 'peephole: #~d ~q -> ~q\n', [Id, From, To]).
debug_compiler(_, _, _, _).
