% FLENG compiler - local analysis

% analyze primitive expressions and collect types + temporaries
%
% builds an environment holding variable information, where the following
% elements are defined:
%
%   temp(VARNUM)        temporary result variable
%   type(VARNUM, TYPE)  variable hold value of known type
%   refs(VARNUM, COUNT) number of variable references in body

analyze((X, Y), Bound1, Bound, Env1, Env) :-
    analyze(X, Bound1, Bound2, Env1, Env2),
    analyze(Y, Bound2, Bound, Env2, Env).
analyze(foreign_call(Term), B1, B, E1, E) :-
    callable(Term),
    Term =.. [_|Args],
    analyze_args(Args, any, Term, B1, B, E1, E).
analyze('$assign'(X, '$VAR'(V)), B, [V|B2], E, [type(V, T)|E3]) :-
    literal_type(X, T, E),
    analyze_arg(X, B, B2, E, E2),
    add_ref(V, E2, E3).
analyze('$address_of'(_, '$VAR'(V)), B, [V|B], E1, [type(V, i)|E]) :-
    add_ref(V, E1, E).
analyze(Term, B1, [R|B], E1, E) :-
    Term = compute(Op, X, Y, '$VAR'(R)),
    computation(Op, [TA, TB, TR]),
    immediate_result_temp(TR, R, [type(R, TR)|E1], E2),
    analyze_args([Op, X, Y], [a, TA, TB], Term, B1, B, E2, E3),
    add_ref(R, E3, E),
    \+memberchk(R, B).
analyze(Term, B1, B, E1, E) :-
    Term = compute(Op, X, Y, R),
    analyze_args([Op, X, Y, R], [a, x, x, x], Term, B1, B, E1, E).
analyze(Term, B1, [R|B], E1, E) :-
    Term = compute(Op, X, '$VAR'(R)),
    computation(Op, [TA, TR]),
    immediate_result_temp(TR, R, [type(R, TR)|E1], E2),
    analyze_args([Op, X], [a, TA], Term, B1, B, E2, E3),
    add_ref(R, E3, E),
    \+memberchk(R, B).
analyze(Term, B1, B, E1, E) :-
    Term = compute(Op, X, R),
    analyze_args([Op, X, R], [a, x, x], Term, B1, B, E1, E).
analyze(Term, B, [R|B], E, [temp(R), type(R, a)|E2]) :-
    Term = '$idle'('$VAR'(R)),
    add_ref(R, E, E2).
analyze(T, Bound1, Bound, Env1, Env) :-
    T =.. [N|Args],
    length(Args, A),
    primitive(N/A, _, R, Types),
    analyze_primitive(R, Args, Types, T, Bound1, Bound, Env1, Env).
analyze(_, B, B, E, E).

analyze_primitive(result, Args, Types, Term, B1, [R|B], E1, E) :-
    split_args(Args, Args2, '$VAR'(R)),
    split_args(Types, _, RT),
    immediate_result_temp(RT, R, [type(R, RT)|E1], E2),
    analyze_args(Args2, Types, Term, B1, B, E2, E),
    \+memberchk(R, B).
analyze_primitive(_, Args, Types, Term, B1, B, E1, E) :-
    analyze_args(Args, Types, Term, B1, B, E1, E).

analyze_args(Args, Types, Term, B1, B, E1, E) :-
    analyze_args(Args, Types, 1, Term, B1, B, E1, E).

analyze_args([], _, _, _, B, B, E, E).
analyze_args(Args, any, Pos, Term, B1, B, E1, E) :-
    analyze_args(Args, [x|any], Pos, Term, B1, B, E1, E).
analyze_args([X|Args], [T|Types], Pos, Term, B1, B, E1, E) :-
    check_type(X, T, Pos, Term, E1),
    !,
    analyze_arg(X, B1, B2, E1, E2),
    Pos2 is Pos + 1,
    analyze_args(Args, Types, Pos2, Term, B2, B, E2, E).

analyze_arg(A, B, B, E, E) :- atomic(A), !.
analyze_arg('$VAR'(I), Bound, [I|Bound], E1, E) :- 
    integer(I), 
    !,
    add_ref(I, E1, E).
analyze_arg([T1|As], Bound1, Bound, E1, E) :-
    !,
    analyze_arg(T1, Bound1, Bound2, E1, E2),
    analyze_arg(As, Bound2, Bound, E2, E).
analyze_arg(T, Bound1, Bound, E1, E) :-
    T =.. [_|Args],
    !,
    analyze_arg(Args, Bound1, Bound, E1, E).
analyze_arg(_, B, B, E, E).

add_ref(V, E1, [refs(V, N2)|E2]) :-
    select(refs(V, N), E1, E2),
    !,
    N2 is N + 1.
add_ref(V, E, [refs(V, 1)|E]).

split_args([R], [], R).
split_args([X|L], [X|L2], R) :-
    split_args(L, L2, R).

literal_type(I, i, _) :- integer(I), !.
literal_type([], a, _) :- !.
literal_type(S, a, _) :- atom(S), !.
literal_type('$VAR'(V), T, E) :-
    memberchk(type(V, T), E),
    !.
literal_type('$VAR'(_), _, _) :- !, fail.
literal_type(T, t, _) :- callable(T), !.

check_type(A, T, Pos, Term, _) :- 
    atom(A),
    check_type2(a, T, Pos, Term).
check_type(N, T, Pos, Term, _) :- 
    integer(N),
    check_type2(i, T, Pos, Term).
check_type('$VAR'(V), T, Pos, Term, E) :- 
    memberchk(type(V, T2), E),
    check_type2(T, T2, Pos, Term).
check_type(_, _, _, _, _).

check_type2(T, T, _, _) :- !.
check_type2(x, _, _, _) :- !.
check_type2(_, x, _, _) :- !.
check_type2(T1, T2, Pos, Term) :-
    type_name(T1, TN1),
    type_name(T2, TN2),
    format(user_error, 'Warning: argument #~d in term `~q\' has type ~a, but was expected to be of type ~a\n', [Pos, Term, TN2, TN1]).

immediate_result_temp(T, V, E1, [temp(V)|E1]) :- immediate_result(T), !.
immediate_result_temp(_, _, E, E).

immediate_result(a).

% tables    

type_name(a, atom).
type_name(i, integer).
type_name(t, tuple).
type_name(c, 'tuple or list').
type_name(l, 'list').
type_name(v, variable).

computation(+, [i, i, i]).
computation(-, [i, i, i]).
computation(*, [i, i, i]).
computation(/, [i, i, i]).
computation(mod, [i, i, i]).
computation(and, [i, i, i]).
computation(or, [i, i, i]).
computation(xor, [i, i, i]).
computation(shl, [i, i, i]).
computation(shr, [i, i, i]).
computation(=, [x, x, a]).
computation(<, [i, i, a]).
computation(>, [i, i, a]).
computation(max, [i, i, i]).
computation(min, [i, i, i]).
computation(sign, [i, i]).
computation(abs, [i, i]).
computation(sqrt, [i, i]).

primitive('$above'/3, above, result, [x, x, a]).
primitive('$assign'/2, assign, none, [x, v]).
primitive('$atomic'/2, atomic, result, [x, a]).
primitive('$below'/3, below, result, [x, x, a]).
primitive('$dbg'/2, debug, none, [a, x]).
primitive('$error'/1, error, none, [a]).
primitive('$unify'/3, unify, result, [x, x, a]).
primitive('$var'/2, test_var, result, [x, a]).
primitive('$remote'/2, test_remote, result, [x, a]).
primitive('$unify_checked'/2, unify_checked, none, [x, x]).
primitive('$number'/2, test_number, result, [x, a]).
primitive('$tuple'/2, test_tuple, result, [x, a]).
primitive('$atom'/2, test_atom, result, [x, a]).
primitive('$match'/3, match, result, [x, x, a]).
primitive('$list'/2, test_list, result, [x, a]).
