; FLENG - Z80 runtime system (CP/M)
;
; register usage:
;
;   A: HL
;   G: IX
;
;   N: NEWGOAL
;   P0-P3: PARAM[<I>]
;   E0-E<n>: ENV[<I>]
;   X0, X1: bc, de
;
; encoding:
;
;   string:             aaaaaaaa aaaaaaa1
;   immediate integer:  0snnnnnn nnnnnnn1
;   cell:               1aaaaaaa aaaaaaa0
;
;   cell = [TAG, CAR, CDR]
;
;   tag:    _TLIVAAA cccccccc
;
;   T = 1: tuple ("AAA" is arity / length - 1)
;   V = 1: var
;   I = 1: integer
;   L = 1: list
;
;   List = [Car, Cdr]
;   Tuple = [Head, Arglist]
;   Integer = [Magnitude, _]
;   Var = [Val|Var, S-list]
;
;   S-List = [[Goal, [Var', Addr], ...], ...]
;
;   Var' = marked var (lowest bit set)
;
; type-ordering: int < string < list < tuple

    org 100h

BASE = 100h

BDOS    = 5
C_WRITESTR = 9
C_WRITE = 2

IMMEDIATE_BIT = 0
LOW_BIT = 7

VAR_BIT = 3
INT_BIT = 4
LIST_BIT = 5
TUPLE_BIT = 6

TIMESLICE = 10

TAG_MASK = 0ff00h
COUNT_MASK = 00ffh
ARITY_MASK = 0700h

MAX_ENV     = 32
HEAP    = 8000h
MAX_CELLS = 0d00h
;MAX_CELLS = 0300h
MAX_GOALS = 400h

    ; entry point
    ld (stackbase), sp
    jp init

    ; extract integer value
    ; A=val -> A (int)
    ; modifies: A, a
fl_getint:
    bit IMMEDIATE_BIT, l
    jr z, l1
    sla h
    sra h
    sra h
    rr l
    ret
l1: ;.
    inc hl
    inc hl
    ld a, (hl)
    inc hl
    ld h, (hl)
    ld l, a
    ret

    ; create integer
    ; A=int -> A
    ; modifies: A, a
fl_mkint:
    ld a, h
    and 0c0h
    jr z,lmki1
    cp 0c0h
    jr nz, l4
lmki1:
    scf           ; immediate integer
    rl l
    rl h
    set LOW_BIT, h
    ret
l4:
    ld bc, 0
fl_mkbigint:
    ld a, 1 << INT_BIT
    ; fall through...
    
    ; allocate cell
    ; A=CAR, a=tag, X0=CDR -> A
    ; modifies: iy, a, X1
fl_cons: 
    push af
    ld de, (freelist)
    ld a, d
    or e
    jr nz, l6
    ld de, ENOCELLS
    jr error
l6:
    ld (consing), de
    ld iy, (consing)
    ld a, (iy)
    ld (freelist), a
    ld a, (iy + 1)
    ld (freelist + 1), a
    pop af
    ld (iy + 1), a
    ld a, 0
    ld (iy), a
    call fl_addref
    ld iy, (consing)
    ld (iy + 2), l
    ld (iy + 3), h
    ld hl, bc
    call fl_addref
    ld iy, (consing)
    ld (iy + 4), l
    ld (iy + 5), h
    ld hl, (cells)
    inc hl
    ld (cells), hl
    push iy
    pop hl
    ret
consing: dw 0

    ; create variable
    ; -> A
op_mkvar:
    call fl_mkvar
    jp next

    ; -> A
fl_mkvar:
    ld a, 1 << VAR_BIT
    ld bc, fl_nil
    ld hl, 0
    call fl_cons
    push hl
    pop iy
    ld (iy + 2), l
    ld (iy + 3), h
    ret

    ; runtime error
    ; de=errmsg
error:
    push ix
    push de
    ld de, errhdr
    ld c, C_WRITESTR
    call BDOS
    pop de
    ld c, C_WRITESTR
    call BDOS
    ld de, errinfo1
    ld c, C_WRITESTR
    call BDOS
    pop ix
    ld l, (ix)
    ld h, (ix + 1)
    call writehex
    ld de, errinfo2
    ld c, C_WRITESTR
    call BDOS
    jp exit
errhdr: db 13, 10, 'Error: $'
errinfo1: db ' [$'
errinfo2: db ']', 13, 10, '$'

    ; write P0 (string) and P1 (hex)
op_dbg:
    push ix
    ld hl, (PARAM0)
    call fl_check_atom
    ld b, 0
    ld c, (hl)
    inc c
    add hl, bc
    ld a, (hl)
    push af
    push hl
    ld a, '$'
    ld (hl), a
    ld de, (PARAM0)
    inc de
    ld c, C_WRITESTR
    call BDOS
    pop hl
    pop af
    ld (hl), a
    ld hl, (PARAM1)
    call writehex
    call newline
    pop ix
    jp next

newline:
    ld de, crnl
    ld c, C_WRITESTR
    jp BDOS
crnl: db 13, 10, '$'

    ; initialize runtime system
fl_initsys:
    ; init sstack
    ld hl, sstack
    ld (sstackp), hl
    ; init goals
    ld ix, goalbuf
    ld b, hi(MAX_GOALS)
    ld de, 10           ; sizeof(goal)
    call linkblocks
    ld (freegoals), hl
    ; init heap
    ld ix, HEAP
    ld b, hi(MAX_CELLS)
    ld de, 6            ; sizeof(cell)
    call linkblocks
    ld (freelist), hl
    ; init queue
    ld hl, queue
    ld (qstart), hl
    ld (qend), hl
    call fl_mkvar
    inc (hl)
    ld (idle), hl
    ret

    ; link sequence of blocks together with ptr in firsts word
    ; de=blocksize, ix=startaddr, b=multiples of 256 -> hl (ptr)
    ; modifies: a, bc, de, hl, ix
linkblocks:
    ld a, 0
    ld (ix), a          ; goal->addr
    ld (ix + 1), a
    push ix
    add ix, de
l100:
    ld c, b
    ld b, a
l5:
    pop hl
    ld (ix), l          ; goal->addr
    ld (ix + 1), h
    push ix
    add ix, de
    djnz l5
    ld b, c
    djnz l100
    pop hl
    ret

    ; increase refcount
    ; A=val -> A
    ; modifies: a
fl_addref:
    ld a, l
    bit IMMEDIATE_BIT, a
    ret nz
    ld a, h
    bit LOW_BIT, a
    ret z
fl_addref0:
    ld a, (hl)
    cp COUNT_MASK
    ret z
    inc (hl)
    ret

    ; decrease refcount
    ; X0=val
    ; modifies: a, iy (must preserve A)
fl_unref:
    ld a, c
    bit IMMEDIATE_BIT, a
    ret nz
    bit LOW_BIT, b
    ret z
    ld a, (bc)
    cp COUNT_MASK
    ret z
    push bc
    pop iy
    dec (iy)
    ret nz
    ; release cell:
    ld a, (iy + 1)          ; save tag
    ld bc, (freelist)
    ld (iy), c
    ld (iy + 1), b
    ld (freelist), iy
    ld bc, (cells)
    dec bc
    ld (cells), bc
    bit INT_BIT, a
    ret nz
    bit VAR_BIT, a          ; variable?
    jr z, l14
    ld (tmp1), iy           ; unbound?
    ld a, (tmp1)
    cp (iy + 2)
    jr nz, l14
    ld a, (tmp1 + 1)
    cp (iy + 3)
    jr nz, l14
    ld c, (iy + 4)             ; bound to itself
    ld b, (iy + 5)
    jr fl_unref
l14:
    ld c, (iy + 4)
    ld b, (iy + 5)
    push bc
    ld c, (iy + 2)
    ld b, (iy + 3)
    call fl_unref
    pop bc
    jp fl_unref

    ; dereference value
    ; A=val -> A
    ; modifies: a, iy
fl_deref:
    bit IMMEDIATE_BIT, l
    ret nz
    push hl
    pop iy
    ld a, (iy + 1)
    bit VAR_BIT, a
    ret z
    ld a, (iy + 2)
    cp l
    jr nz, l17
    ld a, (iy + 3)
    cp h
    ret z
l17:
    ld l, (iy + 2)
    ld h, (iy + 3)
    jr fl_deref

    ; force variables
    ; A=val -> A, CF set to 1 when suspending
    ; modifies: a, iy, X0
fl_force:
    bit IMMEDIATE_BIT, l
    jr nz, clrcarry
    push hl
    pop iy
    ld a, (iy + 1)
    bit VAR_BIT, a
    jr z, clrcarry
    ld a, (iy + 2)
    cp l
    jr nz, lfor1
    ld a, (iy + 3)
    cp h
    jr z, sspush
lfor1:
    ld l, (iy + 2)
    ld h, (iy + 3)
    jr fl_force
sspush:
    ld bc, (sstackp)
    ld a, l
    ld (bc), a
    inc bc
    ld a, h
    ld (bc), a
    inc bc
    ld a, (ix)
    ld (bc), a
    inc bc
    ld a, (ix + 1)
    ld (bc), a
    inc bc
    xor a
    cp c
    jp nz, lfor2
essfull:
    ld de, ESSFULL
    jp error
lfor2:
    ld (sstackp), bc
setcarry:
    scf
    ret
clrcarry:
    and a
    ret

    ; allocate goal
    ; -> N (also in iy)
    ; modifies: A, a, iy, X0, X1
fl_opengoal:
    ld bc, (freegoals)
    ld a, b
    or c
    jr nz, l18
    ld de, ENOGOALS
    jp error
l18:
    ld iy, (freegoals)
    ld a, (iy)
    ld (freegoals), a
    ld a, (iy + 1)
    ld (freegoals + 1), a
    ld hl, (goals)
    inc hl
    ld (goals), hl
    ld (NGOAL), iy
    ; fall through

    ; enqueue goal
    ; N=goal
    ; modifies: A, X1
enqueue:
    ld hl, (qend)
    ld de, (NGOAL)
    ld (hl), e
    inc hl
    ld (hl), d
    inc hl
    ld a, l
    ld de, (qstart)
    cp e
    jr nz, l19
    ld a, h
    cp d
    jr nz, l19a
    ld de, EQUEUEFULL
    jp error
l19: ;.
    ld a, h
l19a:
    cp hi(queuetop)
    jr c, l20
    ld a, l
    cp lo(queuetop)
    jr c, l20
    ld hl, queue ;.
l20:
    ld (qend), hl
    ret

    ; dequeue goal
    ; -> G
    ; modifies: a, A, X0
dequeue:
    ld hl, (qstart)
    push hl
    inc hl
    inc hl
    ld a, h
    cp hi(queuetop)
    jr c, l21
    ld a, l ;.
    cp lo(queuetop)
    jr c, l21
    ld hl, queue
l21:
    ld (qstart), hl
    pop hl
    ld c, (hl)
    inc hl
    ld b, (hl)
    push bc
    pop ix
    ret

    ; prepare for tail call
    ; <AP>=addr -> N
    ; modifies: a, X1, yi
op_tailgoal:
    ld a, (tailcalls)
    ld d, a
    ld a, (timeslice)
    cp d
    jr nz, l70
    call fl_opengoal
    call fetch
    ld (iy), e          ; goal->addr (iy holds N)
    ld (iy + 1), d
    jp next
l70:
    call fetch
    ld (NGOAL), ix
    ld (ix), e          ; goal->addr
    ld (ix + 1), d
    jp next

    ; invoke predicate if timeslice is not exhausted
    ; G=N
    ; modifies: X0, a
op_tailjump:
    ld a, (tailcalls)
    ld b, a
    ld a, (timeslice)
    cp b
    jr nz, l71
    ld a, 0
    ld (tailcalls), a
    jp op_dropgoal
l71:
    ld a, b
    inc a
    ld (tailcalls), a
    ; fall through ...

    ; invoke predicate
    ; N
    ; modifies: A
op_invoke_jump:
    ld sp, (stackbase)
    ld ix, (NGOAL)
    ld e, (ix)          ; goal->addr
    ld d, (ix + 1)
    ld (AP), de
    jp next

    ; prepare invoking predicate
    ; G -> N, A -> addr(N)
    ; modifies: A, X0, N
op_invoke_goal:
    ld (NGOAL), ix
    ld iy, (NGOAL)
    ld a, h
    and 7fh
    ld (iy + 1), a
    ld a, l
    and 0feh
    ld (iy), a
    jp next

    ; N, A=list
op_puttailarg:
    push hl
    pop iy
    ld a, (iy + 4)
    bit IMMEDIATE_BIT, a
    jr z, lpta1
    ld l, (iy + 2)
    ld h, (iy + 3)
lpta1:
    call fl_addref
    ld iy, (NGOAL)
    ld (iy + 8), l
    ld (iy + 9), h
    jp next

    ; drop goal & yield
    ; G
    ; modifies: A
op_dropgoal:
    push ix
    pop hl
    ld bc, (freegoals)
    ld (hl), c
    inc hl
    ld (hl), b
    dec hl
    ld (freegoals), hl
    ld hl, (goals)
    dec hl
    ld (goals), hl
    jp op_yield

    ; does variable have active suspension?
    ; iy: var
    ; modifies: hl, a, de
active:
    ld l, (iy + 4)      ; s-list
    ld h, (iy + 5)
lact0:
    ld a, l
    cp lo(fl_nil)
    jr nz, lact1
    ld a, h
    cp hi(fl_nil)
    jp z, clrcarry
lact1:
    inc hl
    inc hl
    ld e, (hl)      ; check goalptr from CAR for being 0
    inc hl
    ld d, (hl)
    inc hl
    inc de
    inc de
    ld a, (de)
    inc de
    ld b, a
    ld a, (de)
    or b
    jp nz, setcarry
    ld a, (hl)
    inc hl
    ld h, (hl)
    ld l, a
    jp lact0

    ; yield to next process
op_yield:
    ld hl, (ticks)
    inc hl
    ld (ticks), hl
    ld hl, (qstart)
    ld de, (qend)
    ld a, h
    cp d
    jr nz, l22
    ld a, l
    cp e
    jr nz, l22
    ld iy, (idle)       ; queue empty, check idle var for having suspensions
    call active
    jp nc, terminate
    ld iy, (idle)
    ld hl, fl_nil
    call fl_assign
    ld bc, (idle)
    call fl_unref
    ld a, (epoch)
    inc a
    ld (epoch), a
    call fl_mkvar
    inc (hl)
    ld (idle), hl
l22: ;.
    call dequeue
    ; enter goal code
    ; G
    ld hl, (stackbase)
    ld sp, hl
    ld l, (ix)      ; goal->addr
    ld h, (ix + 1)
    ld (AP), hl
    jp next

    ; add goal to pool
    ; X0=addr
    ; modifies: iy
fl_initgoal:
    push bc
    call fl_opengoal
    ld iy, (NGOAL)
    pop bc
    ld (iy), c          ; goal->addr
    ld (iy + 1), b
    ret

terminate:
    ld a, (suspended)
    ld de, EDEADLOCK
    or a
    jp nz, error
    ld bc, (idle)
    call fl_unref
    ld de, ELEAK
    ld a, (cells)
    cp 0
    jp nz, error
    ld a, (cells + 1)
    and a
    jp nz, error
exit:
    ld hl, (stackbase)
    ld sp, hl
    ret

    ; assign value to variable
    ; A=val, iy=var 
    ; modifies: a, X0
fl_assign:
    ld (tmp1), iy
    ld b, h
    ld c, l
    ld a, (iy + 2)
    cp c
    jr nz, lass1
    ld a, (iy + 3) ;.
    cp b
    ret z
lass1:
    ld a, (tmp1)
    cp (iy + 2)
    jr nz, l24
    ld a, (tmp1 + 1)
    cp (iy + 3)
    jr nz, l24
    ld (iy + 2), c
    ld (iy + 3), b
    ld h, b
    ld l, c
    call fl_addref
    ld hl, (tmp1)
    jp resume
l24:
    ld de, EBOUND
    jp error

    ; suspend on variables on sstack & yield
    ; G
    ; modifies: A, a, X0, X1
op_suspend:
    ld hl, fl_nil
    ld (slst), hl
    ld bc, suspend1
    call for_sstack
    push ix
    pop hl
    ld bc, (slst)
    ld a, 1 << LIST_BIT
    call fl_cons           ; slst = cons(goal, slst)
    ld (slst), hl
    ld bc, suspend2
    call for_sstack
    ld hl, sstack
    ld (sstackp), hl
    ld hl, (suspended)
    inc hl
    ld (suspended), hl
    jp op_yield

suspend1:
    ; build slst: A=var, X0=addr
    ld a, l
    or 1            ; mark var
    ld l, a
    ld a, 1 << LIST_BIT
    call fl_cons       ; x = cons(var', addr)
    ld bc, (slst)
    ld a, 1 << LIST_BIT
    call fl_cons       ; slst = cons(x, slst)
    ld (slst), hl
    ret

suspend2:
    ; attach slst to all vars: A=var
    ld de, 4
    add hl, de
    push hl
    ld c, (hl)
    inc hl
    ld b, (hl)
    push bc
    ld hl, (slst)
    ld a, 1 << LIST_BIT
    call fl_cons       ; var->cdr = cons(slst, var->cdr)
    call fl_addref0
    pop bc
    pop iy
    ld (iy), l
    ld (iy + 1), h
    jp fl_unref     ; old var->cdr

slst:   dw 0

    ; resume goals on variable
    ; A=var
    ; modifies:
resume:
    ld (tmp1), hl
    ld iy, (tmp1)
    ld l, (iy + 4)
    ld h, (iy + 5)
    push hl
    ld bc, resume1
    call for_each
    pop bc
    call fl_unref
    ld iy, (tmp1)
    ld a, lo(fl_nil)
    ld (iy + 4), a
    ld a, hi(fl_nil)
    ld (iy + 5), a
    ret
resume1:
    ; check slst of var: A=slst
    ld (slst), hl
    ld iy, (slst)
    ld a, (iy + 2)
    ld b, (iy + 3)
    or b
    ret z
    ld l, (iy + 2)
    ld h, b
    ld (NGOAL), hl
    call enqueue
    ld hl, (suspended)
    dec hl
    ld (suspended), hl
    ld iy, (slst)
    ld a, 0
    ld (iy + 2), a      ; clear goal
    ld (iy + 3), a
    ld l, (iy + 4)
    ld h, (iy + 5)
    ld bc, resume2
    jp for_each
resume2:
    ; find variable/addr pair in slst: A=pair
    push hl
    pop iy
    ld a, (iy + 2)
    and 0feh            ; unmark var
    ld l, a
    ld h, (iy + 3)
    ld a, (tmp1)
    cp l
    ret nz
    ld a, (tmp1 + 1)
    cp h
    ret nz
    ld hl, (NGOAL)
    ld a, (iy + 4)
    ld (hl), a      ; goal->addr
    inc hl
    ld a, (iy + 5)
    ld (hl), a
    jp for_each_abort

    ; match failed
op_fail:
    ld de, EFAIL
    jp error

    ; abort with error
    ; P0=atom, modifies atom buffer
op_abort:
    ld hl, (PARAM0)
    call fl_check_atom
    ld b, 0
    ld c, (hl)
    inc c
    add hl, bc
    ld a, '$'
    ld (hl), a
    ld de, (PARAM0)
    inc de
    jp error

    ; unify with result
    ; P0=val1, P1=val2 -> A (true or false)
    ; modifies: X0, A, iy
op_unify:
    ld hl, (PARAM0)
    ld (up1), hl
    ld hl, (PARAM1)
    ld (up2), hl
    call unify
    jp c, op_false
    jp op_true
    
    ; unify with check
    ; P0=val1, P1=val2
    ; modifies: X0, A, iy
op_unify_checked:
    ld hl, (PARAM0)
    ld (up1), hl
    ld hl, (PARAM1)
    ld (up2), hl
    call unify
    jp nc, next
    ld de, EUNIFY
    jp error

    ; unify result with check
    ; A=val1, iy=val2
    ; modifies: X0, A, iy, a
fl_unify_result:
    ld (up1), hl
    ld (up2), iy
    call unify
    ret nc
    ld de, EUNIFY
    jp error

    ; unify values
    ; up1=val1, up2=val2 -> Carry (1 is mismatch)
    ; modifies: X0, A, iy, a
unify:
    ld hl, (up1)
    call fl_deref
    ld (up1), hl
    ld hl, (up2)
    call fl_deref
    ld (up2), hl
    ld bc, (up1)
    ld a, b
    cp h
    jr nz, l80
    ld a, c
    cp l
    jp z, clrcarry
l80:
    ; X != Y
    bit IMMEDIATE_BIT, c
    jr nz, l82
    push bc
    pop iy
    ld a, (iy + 1)
    bit VAR_BIT, a
    jr z, l82
    call fl_assign
    jp clrcarry
l82:
    ; X is not a variable
    bit IMMEDIATE_BIT, l
    jp nz, setcarry
    ; Y is not immediate
    push hl
    pop iy
    ld a, (iy + 1)
    bit VAR_BIT, a
    jr z, l83
    ld hl, (up1)
    call fl_assign
    jp clrcarry
l83:
    ; Y is not a variable
    bit IMMEDIATE_BIT, c
    jp nz, setcarry
    ; X is not immediate
    ld iy, (up2)
    ld b, (iy + 1)
    ld iy, (up1)
    ld a, (iy + 1)
    cp b
    jp nz, setcarry
    ; tags are equal
    ld l, (iy + 4)
    ld h, (iy + 5)
    push hl
    ld l, (iy + 2)
    ld h, (iy + 3)
    ld (up1), hl
    ld iy, (up2)
    ld l, (iy + 4)
    ld h, (iy + 5)
    push hl
    ld l, (iy + 2)
    ld h, (iy + 3)
    ld (up2), hl
    call unify
    jr nc, l84
    pop hl
    pop hl
    ret
l84:
    pop hl
    ld (up1), hl
    pop hl
    ld (up2), hl
    jp unify
up1: dw 0
up2: dw 0

    ; match values
    ; G, P0=val1, P1=val2 -> A (true, false or nil)
op_match:
    ld hl, (PARAM0)
    ld (up1), hl
    ld hl, (PARAM1)
    ld (up2), hl   
    call match
    jp c, lmat1
    jp z, op_true
    jp op_false
lmat1:
    ld hl, fl_nil
    jp next

    ; match values
    ; G, up1=val1, up2=val2 -> Z + C (Z set if match, C set on suspend)
    ; modifies: X0, a, iy
match:
    ld (ssave), sp
match_rec:
    ld hl, (up1)
    call fl_deref
    ld (up1), hl
    ld hl, (up2)
    call fl_deref
    ld (up2), hl
    ld bc, (up1)
    ld a, b
    cp h
    jr nz, l85
    ld a, c
    cp l
    jp z, mtrue1
l85:
    ; not eq
    bit IMMEDIATE_BIT, c
    jr nz, l86
    push bc
    pop iy
    ld a, (iy + 1)
    bit VAR_BIT, a
    jr z, l86
lmr1:
    call save_and_suspend
    ld sp, (ssave)
    scf
    ret
l86:
    ; X is not a variable
    bit IMMEDIATE_BIT, l
    jp nz, mfalse
    push hl
    pop iy
    ld a, (iy + 1)
    bit VAR_BIT, a
    jr z, l87
    ld b, h
    ld c, l
    jr lmr1
l87:
    ; Y is not a variable
    ld bc, (up1)
    bit IMMEDIATE_BIT, c
    jp nz, mfalse
    ; X and Y are not immediate and not variables
    ld iy, (up2)
    ld b, (iy + 1)
    ld iy, (up1)
    ld a, (iy + 1)
    cp b
    jp nz, mfalse
    ; tags are equal
    ld l, (iy + 4)
    ld h, (iy + 5)
    push hl
    ld l, (iy + 2)
    ld h, (iy + 3)
    ld (up1), hl
    ld iy, (up2)
    ld l, (iy + 4)
    ld h, (iy + 5)
    push hl
    ld l, (iy + 2)
    ld h, (iy + 3)
    ld (up2), hl
    call match_rec
    pop hl
    ld (up2), hl
    pop hl
    ld (up1), hl
    jp match_rec
ssave: dw 0
mtrue:
    ld sp, (ssave)
mtrue1:
    and 0           ; set Z, clear C
    ret
mfalse:
    ld sp, (ssave)
    or 1            ; clear Z, clear C
    ret

save_and_suspend: ; (X0 is var)
    ld hl, (sstackp)
    ld (hl), c
    inc hl
    ld (hl), b
    inc hl
    ld a, (ix)
    ld (hl), a
    inc hl
    ld a, (ix + 1)
    ld (hl), a
    inc hl
    ld (sstackp), hl
    ld a, l
    or h
    jp z, essfull
    scf
    ret

    ; compare for equality
    ; P0, P1 -> Carry (set if not equal)
    ; modifies: a, X0
fl_eq:
    ld bc, (PARAM0)
    ld a, (PARAM1)
    cp c
    jp nz, setcarry
    ld a, (PARAM1 + 1)
    cp b
    jp nz, setcarry
    jp clrcarry

    ; A=val -> A (int)
op_intval:
    call fl_getint
    jp next

    ; A=int -> A
op_to_int:
    call fl_mkint
    jp next

    ; arithmetic
    ; (SP) + A -> A (int)
op_add:
    pop bc
    add hl, bc
    jp next

    ; (SP) - A -> A (int)
op_sub:
    pop bc
    call sub16
    jp next

    ; (SP) * A -> A (int)
op_mul:
    pop bc
    ld d, h
    ld e, l
    ; http://map.grauw.nl/articles/mult_div_shifts.php
    ; HL = BC * DE
    ld a,b
    ld b,16
Mult16_Loop:
    add hl,hl
    sla c
    rla
    jr nc,Mult16_NoAdd
    add hl,de
Mult16_NoAdd:
    djnz Mult16_Loop
    jp next

    ; (SP) / A -> A (int)
op_div:
    ld (up2), hl
    pop hl
    ld (up1), hl
    ld a, 0
    ld (divsign), a
    call divmod
    ld hl, bc
    ld a, (divsign)
    or a
    call nz, negate
    jp next

    ; (SP) % A -> A (int)
op_mod:
    ld (up2), hl
    pop hl
    ld (up1), hl
    call divmod
    jp next

divsign: db 0
divmod:
    ld hl, (up1)
    ld d, h
    ld a, h
    and 80h
    call nz, negate
    ld b, h
    ld c, l
    ld hl, (up2)
    ld e, h
    ld a, h
    and 80h
    call nz, negate
    ld a, d
    xor e
    jp p, ldiv1
    ld a, 1
    ld (divsign), a
ldiv1:
    xor a
    cp h
    jr nz, l120
    ld a, l
    or a
    jr nz, l120
    ld de, EDIV0
    jp error
l120:
    ld d, h
    ld e, l
    ; http://map.grauw.nl/articles/mult_div_shifts.php
    ; BC + HL = BC / DE
Div16:
    ld hl,0
    ld a,b
    ld b,8
Div16_Loop1:
    rla
    adc hl,hl
    sbc hl,de
    jr nc,Div16_NoAdd1
    add hl,de
Div16_NoAdd1:
    djnz Div16_Loop1
    rla
    cpl
    ld b,a
    ld a,c
    ld c,b
    ld b,8
Div16_Loop2:
    rla
    adc hl,hl
    sbc hl,de
    jr nc,Div16_NoAdd2
    add hl,de
Div16_NoAdd2:
    djnz Div16_Loop2
    rla
    cpl
    ld b,c
    ld c,a
    ret

    ; HL = BC - HL
sub16:
    ld a, l
    cpl
    ld l, a
    ld a, h
    cpl
    ld h, a
    inc hl
    add hl, bc
    ret

    ; integer square root
    ; A (int) -> A (int)
op_sqrt:
    ; http://map.grauw.nl/sources/external/z80bits.html#3.2
    ld a, -1
    ld d, a
    ld e, a
Sqr16_Loop:
    add hl,de
    inc a
    dec e
    dec de
    jr c,Sqr16_Loop
    ld h, 0
    ld l, a
    jp next

    ; binary and
    ; (SP) & A (int) -> A (int)
op_and:
    pop bc
    ld a, l
    and c
    ld l, a
    ld a, h
    and b
    ld h, a
    jp next

    ; binary or
    ; (SP) | A (int) -> A (int)
op_or:
    pop bc
    ld a, l
    or c
    ld l, a
    ld a, h
    or b
    ld h, a
    jp next

    ; binary xor
    ; (SP) ^ A (int) -> A (int)
op_xor:
    pop bc
    ld a, l
    xor c
    ld l, a
    ld a, h
    xor b
    ld h, a
    jp next

    ; binary shift left
    ; (SP) << A (int) -> A (int)
op_shl:
    ld a, l
    and a
    jr nz, lshl1
    pop hl
    jp next
lshl1:
    ex (sp), hl
    pop bc
    ld b, c
lshl2:
    sla l
    rl h
    djnz lshl2
    jp next

    ; binary shift right
    ; (SP) >> A (int) -> A (int)
op_shr:
    ld a, l
    and a
    jr nz, lshr1
    pop hl
    jp next
lshr1:
    ex (sp), hl
    pop bc
    ld b, c
lshr2:
    sra h
    rr l
    djnz lshr2
    jp next

    ; match atom or small int
    ; A=val, <AP>=atomic + goaladdr + jumpaddr
    ; modifies: a
op_isval:
    call fetch
    ld a, d
    cp h
    jp nz, trynext
    ld a, e
    cp l
    jp nz, trynext
    ; fall through...
skip2:
    call skip
    jp op_skip

    ; match boxed integer
    ; A=val, <AP>=int + goaladdr + jumpaddr
    ; modifies: a
op_isint:
    call fetch
    bit IMMEDIATE_BIT, l
    jp nz, trynext
    push hl
    pop iy
    ld a, (iy + 1)
    bit INT_BIT, a
    jp z, trynext
    ld a, (iy + 2)
    cp e
    jp nz, trynext
    ld a, (iy + 3) ;.
    cp d
    jp nz, trynext
    jp skip2

    ; check value for being false and jump if it isn't
    ; A=val, <AP>=jumpaddr
    ; modifies: a
op_isfalse:
    ld a, l
    cp lo(fl_false)
    jp nz, op_jump
    ld a, h
    cp hi(fl_false)
    jp nz, op_jump
    ; fall through

op_skip:
    call skip
    jp next

    ; check value for being true and jump if it isn't
    ; A=val, <AP>(jumpaddr)
    ; modifies: a
op_istrue:
    ld a, l
    cp lo(fl_true)
    jp nz, op_jump
    ld a, h
    cp hi(fl_true)
    jp z, op_skip
    jp op_jump

    ; check value for being atomic (small int or atom)
    ; A=val, <AP>=addr
    ; modifies: a
op_isatomic:
    and a
    bit IMMEDIATE_BIT, l
    jp z, op_jump
    jp op_skip

    ; check value for being list
    ; A=val -> C (set if not list)
    ; modifies: a
fl_islist_t:
    bit IMMEDIATE_BIT, l
    jp nz, setcarry
    inc hl
    ld a, (hl)
    dec hl
    and a           ; clear C
    bit LIST_BIT, a
    ret nz
    jp setcarry

    ; check value for being list
    ; A=val, <AP>=jumpaddr
op_islist_t:
    call fl_islist_t
    jp c, op_jump
    jp op_skip

    ; check value for being list and deconstruct
    ; A=val, <AP>=jumpaddr
op_islist_n:
    call fl_islist_t
    jp c, op_jump
    call skip
    jp op_decons

    ; check value for being list and deconstruct
    ; A=val, <AP>=goaladdr + jumpaddr -> A(car), cdr on stack
op_islist:
    call fl_islist_t
    jp c, trynext
    call skip
    call skip
    ; fall through ...

    ; deconstruct cell
    ; A=val -> A(car), cdr on stack
op_decons:
    push hl
    pop iy
    ld e, (iy + 4)
    ld d, (iy + 5)
    push de
    ld l, (iy + 2)
    ld h, (iy + 3)
    jp next

    ; check value for being tuple and deconstruct
    ; A=val, <AP>=arity + goaladdr + jumpaddr -> A (head), arglist on stack
op_istuple:
    call fetch
    push de
    bit IMMEDIATE_BIT, l
    jp nz, trynext
    inc hl
    ld a, (hl)
    dec hl
    bit TUPLE_BIT, a
    jp z, trynext
    and ARITY_MASK >> 8
    pop bc
    cp c
    jp nz, trynext
    call skip
    call skip
    jp op_decons

    ; check for numeric or immediate equality
    ; P0=val1, P1=val2 -> A
    ; modifies: X0, iy
op_equal:
    ld a, (PARAM0)
    ld hl, (PARAM1)
    cp l
    jr nz, leql1
    ld a, (PARAM0 + 1)
    cp h
    jp z, op_true
leql1:
    ld a, (PARAM0)
    bit IMMEDIATE_BIT, a
    jp nz, op_false    
    bit IMMEDIATE_BIT, l
    jp nz, op_false
    ld iy, (PARAM0)
    ld b, (iy + 1)
    ld iy, (PARAM1)
    ld a, (iy + 1)
    cp b
    jp nz, op_false
    ld iy, (PARAM0)
    ld hl, (PARAM1)
    inc hl
    inc hl
    ld a, (hl)
    cp (iy + 2)
    jp nz, op_false
    inc hl
    ld a, (hl)
    cp (iy + 3)
    jp nz, op_false
    jp op_true

    ; check for numerically less-than
    ; P0=val1, P1=val2 -> A
    ; modifies: X0, iy, P0, P1
op_less:
    ld hl, (PARAM0)
    call fl_getint
    ld (up1), hl
    ld hl, (PARAM1)
    call fl_getint
    ld (up2), hl
    call less
    jp c, op_true
    jp op_false
 
    ; check for numerically greater-than
    ; P0=val1, P1=val2 -> A
    ; modifies: X0, iy, P0, P1
op_greater:
    ld hl, (PARAM0)
    call fl_getint
    ld (up2), hl
    ld hl, (PARAM1)
    call fl_getint
    ld (up1), hl
    call less
    jp c, op_true
    jp op_false

    ; from "Programming the Z80", Rodney Zacks, p.531
    ; up1=int1, up2=int2 -> C (set if less)
    ; modifies: a, iy
less:
    ld iy, up2
    ld a, (up1 + 1)
    and 80h
    jp nz, lles1
    bit 7, (iy + 1)
    ret nz
    ld a, (up1 + 1)
    cp (iy + 1)
    ret nz
    ld a, (up1)
    cp (iy)
    ret
lles1:
    xor (iy + 1)
    rla
    ret c
    ld a, (up1 + 1)
    cp (iy + 1)
    ret nz
    ld a, (up1)
    cp (iy)
    ret

    ; update goal-ptr and continue at new address
    ; <AP>=addr + jumpaddr, G
    ; modifies: X0
trynext:
    call fetch
    ld (ix), e
    ld (ix + 1), d
    jp op_jump

    ; type check for integer
    ; A=val
    ; modifies: a, iy
fl_check_integer:
    bit IMMEDIATE_BIT, l
    jr z, lchki1
    bit LOW_BIT, h
    ret nz
lchki1:
    push hl
    pop iy
    ld a, (iy + 1)
    bit INT_BIT, a
    ret nz
    ld de, ENOINT
    jp error

    ; type check for string
    ; A=val
    ; modifies: a
fl_check_atom:
    bit IMMEDIATE_BIT, l
    jr z, lchka1
    bit LOW_BIT, h
    ret z
lchka1: ;.
    ld de, ENOSTRING
    jp error

    ; type check for cell
    ; A=val
fl_check_cell:
    bit IMMEDIATE_BIT, l
    ret z
    ld de, ENOCELL
    jp error

    ; type check for tuple
    ; A=val
    ; modifies: a, iy
fl_check_tuple:
    bit IMMEDIATE_BIT, l
    jr nz, lchkt1
    push hl
    pop iy
    ld a, (iy + 1)
    bit TUPLE_BIT, a
    ret nz
lchkt1:
    ld de, ENOTUPLE
    jp error

    ; type check for list
    ; A=val
    ; modifies: a, iy
fl_check_list:
    bit IMMEDIATE_BIT, l
    jr nz, lchkl1
    push hl
    pop iy
    ld a, (iy + 1)
    bit LIST_BIT, a
    ret nz
lchkl1:
    ld de, ENOLIST
    jp error

    ; type check for var
    ; A=val
    ; modifies: a, iy
fl_check_var:
    bit IMMEDIATE_BIT, l
    jr nz, lchkv1
    push hl
    pop iy
    ld a, (iy + 1)
    bit VAR_BIT, a
    ret nz
lchkv1:
    ld de, ENOVAR
    jp error

    ; test if tuple
    ; A=val, <AP>=addr
op_istuple_t:
    bit IMMEDIATE_BIT, l
    jp nz, op_jump
    inc hl
    ld a, (hl)
    dec hl
    bit TUPLE_BIT, a
    jp z, op_jump
    jp op_skip

    ; test if int
    ; P0=val -> A (bool)
op_test_number:
    ld hl, (PARAM0)
    bit IMMEDIATE_BIT, l
    jp nz, lti1
    inc hl
    ld a, (hl)
    bit INT_BIT, a
    jp nz, op_true
    jp op_false
lti1: ;.
    bit LOW_BIT, h
    jp nz, op_true
    jp op_false

    ; test if atom
    ; P0=val -> A (bool)
op_test_atom:
    ld hl, (PARAM0)
    bit IMMEDIATE_BIT, l
    jp z, op_false
    bit LOW_BIT, h
    jp nz, op_false
    jp op_true

    ; test if var
    ; P0=val -> A (bool)
op_test_var:
    ld hl, (PARAM0)
    bit IMMEDIATE_BIT, l
    jp nz, op_false
    inc hl
    ld a, (hl)
    bit VAR_BIT, a
    jp nz, op_true
    jp op_false

    ; test if tuple
    ; P0=val -> A (bool)
op_test_tuple:
    ld hl, (PARAM0)
    bit IMMEDIATE_BIT, l
    jp nz, op_false
    inc hl
    ld a, (hl)
    bit TUPLE_BIT, a
    jp nz, op_true
    jp op_false

    ; test if list
    ; P0=val -> A (bool)
op_test_list:
    ld hl, (PARAM0)
    bit IMMEDIATE_BIT, l
    jp nz, op_false
    inc hl
    ld a, (hl)
    bit LIST_BIT, a
    jp nz, op_true
    jp op_false

    ; absolute value
    ; A (int) -> A (int)
op_abs:
    bit 7, h
    jp z, next
    call negate
    jp next

    ; sign of integer
    ; A=val (int) -> A (int)
    ; modifies: a
op_sign:
    ld a, h
    or l
    jp nz, lsign1
    ld hl, 0
    jp next
lsign1:
    bit 7, h
    jr z, lsign2
    ld hl, 0ffffh
    jp next
lsign2:
    ld hl, 1
    jp next

    ; ordered compare
    ; P0=val1, P1=val2 -> A (bool or nil)
op_below:
    ld hl, (PARAM0)
    ld (up1), hl
    ld hl, (PARAM1)
    ld (up2), hl
    call ordering
    jr c, ofail
    jp m, op_true
    jp op_false
ofail:
    ld hl, fl_nil
    jp next

    ; ordered compare
    ; P0=val1, P1=val2 -> A (bool or nil)
op_above:
    ld hl, (PARAM0)
    ld (up1), hl
    ld hl, (PARAM1)
    ld (up2), hl
    call ordering
    jr c, ofail
    jp m, op_false
    jp z, op_false
    jp op_true

    ; access idle var
    ; <AP>=cache -> A (bool)
    ; modifies: X0, a
op_idle:
    call fetcha
    ld b, (hl)
    inc b
    ld a, (epoch)
    cp b
    jr nz, lidl1
    xor a
    ld (hl), a
    jp op_true
lidl1:
    ld (hl), a
    ld hl, (idle)
    call sspush
    jp op_false

    ; maximum value
    ; max((SP), A) -> A (int)
op_max:
    pop bc
    ld (up1), bc
    ld (up2), hl
    ld hl, (up2)
    call less
    jp c, next
    ld hl, (up1)
    jp next

    ; minimum value
    ; min((SP), A) -> A (int)
op_min:
    pop bc
    ld (up1), bc
    ld (up2), hl
    ld hl, (up1)
    call less
    jp c, next
    ld hl, (up2)
    jp next

    ; allocate list cell
    ; A=cdr, (sp + 2)=car -> A
op_mklist:
    ld b, h
    ld c, l
    pop hl
    ld a, 1 << LIST_BIT
    call fl_cons
    jp next

    ; allocate tuple
    ; A=args, <AP>=atom + arity -> A
op_mktuple:
    push hl
    call fetch
    push de
    call fetch
    ld a, e
    or 1 << TUPLE_BIT
    pop hl
    pop bc
    call fl_cons
    jp next

    ; mutate list into general tuple
    ; A=list, a=arity -> A
fl_mktuple0:
    ld b, a
    and ~(ARITY_MASK >> 8)
    jr z, lmkt1
    ld de, EARITY
    jp error
lmkt1:
    ld a, b
    inc hl    
    and ARITY_MASK >> 8
    or 1 << TUPLE_BIT
    ld (hl), a
    dec hl
    ret

op_mktuple0:
    call fetch
    ld a, e
    call fl_mktuple0
    jp next

    ; threaded interpreter
next:
    call fetch
    push de
    ret

    ; -> de
    ; modifies: AP, bc, a
fetch:
    ld bc, (AP)
    ld a, (bc)
    ld e, a
    inc bc
    ld a, (bc)
    ld d, a
    inc bc
    ld (AP), bc
    ret

    ; -> A
    ; modifies: AP, bc, a
fetcha:
    call fetch
    ld l, e
    ld h, d
    ret

skip:
    ld bc, (AP)
    inc bc
    inc bc
    ld (AP), bc
    ret

    ; sp: points to threaded code
fl_exec:
    pop bc
    ld (AP), bc
    jr next

    ; unconditional addref, expects block
    ; A=val
op_addref:
    inc (hl)
    jp next

    ; A=val
op_addrefx:
    call fl_addref
    jp next

    ; P0=val, A=var
op_assign:
    push hl
    pop iy
    ld hl, (PARAM0)
    call fl_assign
    jp next

    ; test if atomic
    ; P0=val -> A (bool)
op_atomic:
    ld hl, (PARAM0)
    bit IMMEDIATE_BIT, l
    jp nz, op_true
    inc hl
    ld a, (hl)
    bit INT_BIT, a
    jp nz, op_true
    ; fall through

    ; -> A
op_false:
    ld hl, fl_false
    jp next

    ; -> A
op_true:
    ld hl, fl_true
    jp next

    ; <AP>=addr -> A
peek1:
    call fetch
    ld a, (de)
    ld l, a
    inc de
    ld a, (de)
    ld h, a
    ret

    ; <AP>=addr -> A
op_peek:
    call peek1
    jp next

    ; <AP>=addr -> A
op_peek_deref:
    call peek1
    jp op_deref

    ; A, <AP>=addr
op_poke:
    call fetch
    ld a, l
    ld (de), a
    inc de
    ld a, h
    ld (de), a
    jp next

    ; A, <AP>=addr
op_store:
    call fl_addref
    jr op_poke

    ; <AP>=addr, mkvar -> A
op_storevar:
    call fl_mkvar
    jr op_store

    ; A=val
op_check_atom:
    call fl_check_atom
    jp next
op_check_cell:
    call fl_check_cell
    jp next
op_check_integer:
    call fl_check_integer
    jp next
op_check_list:
    call fl_check_list
    jp next
op_check_tuple:
    call fl_check_tuple
    jp next
op_check_var:
    call fl_check_var
    jp next

op_intval_checked:
    call fl_check_integer
    jp op_intval

op_tryelse:
    ld sp, (stackbase)
    jp next

    ; <AP> -> A
op_lit:
    call fetcha
    jp next

    ; A -> A
op_deref:
    call fl_deref
    jp next

    ; A -> A, <AP>=jumpaddr
op_force:
    call fl_force
    jr c, op_jump
    jp op_skip

    ; <AP>=jumpaddr
op_forcearg:
    ld l, (ix + 2)
    ld h, (ix + 3)
    call fl_force
    jr c, op_jump
    ld a, (ix + 2)
    cp l
    jr nz, lfoa1
    ld a, (ix + 3)
    cp h
    jp z, op_skip
lfoa1:
    call fl_addref
    push hl
    ld c, (ix + 2)
    ld b, (ix + 3)
    call fl_unref
    pop hl
    ld (ix + 2), l
    ld (ix + 3), h
    jp op_skip
    
    ; <AP>=addr
op_jump:
    call fetch
    ld (AP), de
    jp next

    ; <AP>=addr
op_fcall:
    call fetch
    ld bc, next
    push bc
    push de
    ret
    
    ; -> A
op_getarg1:
    ld l, (ix + 2)
    ld h, (ix + 3)
    jp next
op_getarg2:
    ld l, (ix + 4)
    ld h, (ix + 5)
    jp next
op_getarg3:
    ld l, (ix + 6)
    ld h, (ix + 7)
    jp next
op_getarg4:
    ld l, (ix + 8)
    ld h, (ix + 9)
    jp next

    ; A -> A<n>
op_putarg1:
    ld (iy + 2), l
    ld (iy + 3), h
    jp next
op_putarg2:
    ld (iy + 4), l
    ld (iy + 5), h
    jp next
op_putarg3:
    ld (iy + 6), l
    ld (iy + 7), h
    jp next
op_putarg4:
    ld (iy + 8), l
    ld (iy + 9), h
    jp next

    ; A -> A<n>
op_storearg1:
    call fl_addref
    ld (iy + 2), l
    ld (iy + 3), h
    jp next
op_storearg2:
    call fl_addref
    ld (iy + 4), l
    ld (iy + 5), h
    jp next
op_storearg3:
    call fl_addref
    ld (iy + 6), l
    ld (iy + 7), h
    jp next
op_storearg4:
    call fl_addref
    ld (iy + 8), l
    ld (iy + 9), h
    jp next

    ; A -> P<n>
op_putparam1:
    ld (PARAM0), hl
    jp next
op_putparam2:
    ld (PARAM1), hl
    jp next
op_putparam3:
    ld (PARAM2), hl
    jp next
op_putparam4:
    ld (PARAM3), hl
    jp next

    ; -> iy(N)
op_ngoal:
    ld iy, (NGOAL)
    jp next

    ; <AP>=addr
op_not_suspending:
    ld a, (sstackp)
    or a
    jp z, op_jump
    jp op_skip
op_suspending:
    ld a, (sstackp)
    or a
    jp nz, op_jump
    jp op_skip

    ; P0, P1, <AP>=jumpaddr
op_guard:
    call fl_eq
    jp c, op_jump
    jp op_skip

    ; <AP>=addr
op_opengoal:
    call fl_opengoal
    call fetch
    ld (iy), e          ; goal->addr
    ld (iy + 1), d
    jp next

    ; -> A
op_pop:
    pop hl
    jp next

    ; A
op_push:
    push hl
    jp next

op_reset:
    ld bc, sstack
    ld (sstackp), bc
    jp next

    ; <AP>=addr
op_drop:
    call fetch
    ld a, (de)
    ld c, a
    inc de
    ld a, (de)
    ld b, a
    call fl_unref
    jp next

op_droparg1:
    ld c, (ix + 2)
    ld b, (ix + 3)
    call fl_unref
    jp next
op_droparg2:
    ld c, (ix + 4)
    ld b, (ix + 5)
    call fl_unref
    jp next
op_droparg3:
    ld c, (ix + 6)
    ld b, (ix + 7)
    call fl_unref
    jp next
op_droparg4:
    ld c, (ix + 8)
    ld b, (ix + 9)
    call fl_unref
    jp next

    ; E<I> -> Arg<J> -> A, <AP>=[I,J] (already multiplied by 2)
    ; uses N
    ; modifies: bc, de, a
op_move_e_a:
    call fetch
    push de
    ld hl, ENV
    ld a, d
    ld de, 0
    ld e, a
    add hl, de
    ld c, (hl)
    inc hl
    ld b, (hl)
    pop de
    ld a, 0
    ld d, a
    ld iy, (NGOAL)
    add iy, de
    ld hl, bc
    ld (iy + 2), l
    ld (iy + 3), h
    ld iy, (NGOAL)
    jp op_addrefx

    ; mkvar -> E<I> -> Arg<J> -> A, <AP>=[I,J] (already multiplied by 2)
    ; uses: N
    ; modifies: bc, de, a
op_move_var_e_a:
    call fl_mkvar
    inc (hl)        ; addref
    push hl
    call fetch
    push de
    ld hl, ENV
    ld a, d
    ld de, 0
    ld e, a
    add hl, de
    pop de
    pop bc
    ld (hl), c
    inc hl
    ld (hl), b
    ld a, 0
    ld d, a
    ld iy, (NGOAL)
    add iy, de
    ld hl, bc
    ld (iy + 2), l
    ld (iy + 3), h
    ld iy, (NGOAL)
    jp op_addrefx

    ; Arg<I> -> E<J> -> A, <AP>=[I,J] (already multiplied by 2)
    ; uses: G
    ; modifies: bc, de, a
op_move_a_e:
    push ix
    call fetch
    push de
    ld a, d
    ld de, 0
    ld e, a
    add ix, de
    ld c, (ix + 2)
    ld b, (ix + 3)
    pop de
    ld a, 0
    ld d, a
    ld hl, ENV
    add hl, de
    ld (hl), c
    inc hl
    ld (hl), b
    pop ix
    ld hl, bc
    jp op_addrefx

    ; E<I> -> P<J> -> A, <AP>=[I,J] (already multiplied by 2)
    ; modifies: bc, de, a
op_move_e_p:
    call fetch
    push de
    ld a, d
    ld de, 0
    ld e, a
    ld hl, ENV
    add hl, de
    ld c, (hl)
    inc hl
    ld b, (hl)
    ld hl, bc
    call fl_deref
    pop de
    ld bc, hl
    ld hl, PARAM0
    ld a, 0
    ld d, a
    add hl, de
    ld (hl), c
    inc hl
    ld (hl), b
    ld hl, bc
    jp next

    ; mkvar -> E<I> -> P<J> -> A, <AP>=[I,J] (already multiplied by 2)
    ; modifies: bc, de, a
op_move_var_e_p:
    call fl_mkvar
    inc (hl)            ; addref
    push hl
    call fetch
    push de
    ld a, d
    ld de, 0
    ld e, a
    ld hl, ENV
    add hl, de
    pop de
    pop bc
    ld (hl), c
    inc hl
    ld (hl), b
    ld hl, PARAM0
    ld a, 0
    ld d, a
    add hl, de
    ld (hl), c
    inc hl
    ld (hl), b
    ld hl, bc
    jp next

    ; auxilliary routines

    ; ordering of types 
    ; A=val -> b (order code)
    ; modifies: A, a
type_order:
    bit IMMEDIATE_BIT, l
    jr z, lord1
    bit LOW_BIT, h
    jr z, lord2
    ld b, 1     ; int
    ret
lord2:
    ld b, 2     ; str
    ret
lord1:
    inc hl
    ld a, (hl)
    bit VAR_BIT, a
    jr z, lord3
    ld b, 0     ; var
    ret
lord3:
    bit TUPLE_BIT, a
    jr z, lord4
    ld b, 4     ; tup
    ret
lord4:
    bit INT_BIT, a
    jr z, lord5
    ld b, 1     ; int
    ret
lord5:
    ld b, 3     ; list
    ret

    ; ordering of values
    ; up1=val1, up2=val2 -> Z, S, C (set if P0 == P1, P0 < P1, C: suspend)
    ; modifies: a, A, X0, iy
ordering:
    ld (ssave), sp
order_rec:
    ld hl, (up1)
    call type_order
    ld a, b
    cp 0
    jr nz, lorc1
    ld bc, (up1)
lorc2:
    call save_and_suspend
    scf
    ret
lorc1:
    ld c, b
    ld hl, (up2)
    call type_order
    ld a, b
    cp 0
    jr nz, lorc3
    ld bc, (up2)
    jr lorc2
lorc3:
    ld a, c
    sub b
    jr z, lorc4
    scf
    ccf
    ret
lorc4:
    ld iy, cmptable
    sla b
    ld c, b
    ld b, 0
    add iy, bc
    ld l, (iy)
    ld h, (iy + 1)
    jp (hl)
cmptable:
    dw cmpptr
    dw cmpint
    dw cmpstr
    dw cmplst
    dw cmptup
cmpint:
    ld hl, (up1)
    call fl_getint
    ld b, h
    ld c, l
    ld hl, (up2)
    call fl_getint
    call sub16
    ld a, l
    or h
    ret
cmpptr:
    ld a, (up2)
    ld b, a
    ld a, (up1)
    sub b
    ld a, (up2 + 1)
    ld b, a
    ld a, (up1 + 1)
    sbc a, b
    ret
cmplst:
    ld iy, (up1)
    push iy
    ld a, (iy + 2)
    ld (up1), a
    ld a, (iy + 3)
    ld (up1 + 1), a
    ld iy, (up2)
    push iy
    ld a, (iy + 2)
    ld (up2), a
    ld a, (iy + 3)
    ld (up2 + 1), a
    call order_rec
    pop hl
    pop iy
    scf
    ccf
    ret nz
    ld a, (iy + 4)
    ld (up1), a
    ld a, (iy + 5)
    ld (up1 + 1), a
    ld de, 4
    add hl, de
    ld a, (hl)
    ld (up2), a
    inc hl
    ld a, (hl)
    ld (up2 + 1), a
    jp order_rec
cmptup:
    ld iy, (up1)
    ld a, (iy + 1)
    ld iy, (up2)
    ld b, (iy + 1)
    sub b
    scf
    ccf
    ret nz
    jp cmplst
cmpstr:
    ld iy, (up1)
    ld c, (iy)
    ld b, 0
    inc iy
    ld hl, (up2)
    inc hl
strl1:
    ld a, (iy)
    inc iy
    cpi
    ret nz
    djnz strl1
    scf
    ccf
    ret

    ; iterate over list
    ; A=list, X0=addr, invoke addr with element in A
    ; modifies: A, a, iy, X0
for_each:
    ld iy, (for_addr)
    push iy
    ld (for_addr), bc
l30:
    ld a, lo(fl_nil)
    cp l
    jr nz, l31
    ld a, hi(fl_nil)
    cp h
    jr nz, l31
    pop iy
    ld (for_addr), iy
    ret
l31:
    push hl
    pop iy
    ld l, (iy + 2)
    ld h, (iy + 3)
    push iy         ; B
    ld bc, l32
    push bc         ; A
    ld iy, (for_addr)
    jp (iy)
l32: ;.
    pop iy          ; B
    ld l, (iy + 4)
    ld h, (iy + 5)
    jr l30
for_addr: dw 0
for_each_abort:
    pop bc          ; A
    pop bc          ; B
    pop bc          ; old for_addr
    ld (for_addr), bc
    ret

    ; iterate over sstack
    ; X0=addr, invoke addr with entries in A(var), X0(saddr)
    ; modifies: A, a, iy, X0, X1
    ; (assumes sstack is not empty)
for_sstack:
    ld iy, (for_addr)
    push iy
    ld (for_addr), bc
    ld iy, sstack
    ld a, (sstackp)     ; low byte
    ld b, a
    srl b
    srl b
l40:
    ld l, (iy)
    ld h, (iy + 1)
    push bc     ; A
    ld c, (iy + 2)
    ld b, (iy + 3)
    push iy     ; B
    ld de, l41
    push de
    ld iy, (for_addr)
    jp (iy)
l41:
    pop iy      ; B
    pop bc      ; A
    ld de, 4
    add iy, de
    djnz l40
    pop hl
    ld (for_addr), hl
    ret

    ; write hex number
    ; A=num
    ; modifies: a, ...
writehex:
    push hl
    ld a, h
    call wbyte
    pop hl
    ld a, l
wbyte:
    push af
    srl a
    srl a
    srl a
    srl a
    call wdigit
    pop af
    and 0fh
wdigit:
    cp 10
    jr c, l110
    add 'A' - '0' - 10
l110:
    add '0'
    ld e, a
    ld c, C_WRITE
    jp BDOS

    ; A, <AP> = [Count, Val, Label, ...]
op_lookup:
    call fetch
llok1:
    push de
    call fetch
    ld a, e
    cp l
    jr nz, llok2
    ld a, d
    cp h
    jr nz, llok2
    pop de
    jp op_jump
llok2:
    call skip
    pop de
    dec e
    jr nz, llok1
    jp op_jump

    ; iy=ENV-ptr, A
fl_xassign:
    push hl
    ld l, (iy)
    ld h, (iy + 1)
    push hl
    pop iy
    pop hl
    jp fl_assign

    ; foreign_call wrappers

    ; convert integer to list
    ; P0=num, P1=list, P2=tail
fl_number_to_list_3:
    ld hl, (PARAM0)
    call fl_getint
    ld de, ln2lbuf
    ld a, l
    or h
    jr nz, ln2l1
    ld a, '0'
    ld (ln2lbuf), a
    inc de
    jp consbuf
ln2l1:
    bit 7, h
    jr z, ln2l2
    ld a, '-'
    ld (ln2lbuf), a
    inc de
    call negate
ln2l2:
    call Num2Dec
consbuf:
    ld hl, (PARAM2)
ln2l3:
    push hl
    dec de
    ld a, (de)
    add a, a
    or 1
    ld l, a
    ld h, 80h
    ld a, 1 << LIST_BIT
    pop bc
    push de
    call fl_cons
    pop de
    ld a, e
    cp lo(ln2lbuf)
    jr nz, ln2l3
    ld iy, (PARAM1)
    jp fl_unify_result
ln2lbuf: ds 8

    ; A=int -> -A
    ; https://www.smspower.org/Development/Z80ProgrammingTechniques#x16bitNeg
negate:
 xor a
 sub l
 ld l,a
 sbc a,a
 sub h
 ld h,a
    ret

    ; http://map.grauw.nl/sources/external/z80bits.html#5.1
    ; Input: HL = number to convert, DE = location of ASCII string
    ; Output: ASCII string at (DE)
Num2Dec:
    ld (tmp1), de
	ld	bc,-10000
	call	Num1
    call n2dsk0
	ld	bc,-1000
	call	Num1
    call n2dsk0
	ld	bc,-100
	call	Num1
    call n2dsk0
	ld	c,-10
	call	Num1
    call n2dsk0
	ld	c,b
Num1:	ld	a,'0'-1
Num2:	inc	a
	add	hl,bc
	jr	c,Num2
	sbc	hl,bc
	ld	(de),a
	inc	de
	ret
n2dsk0:
    cp '0'
    ret nz
    ld a, (tmp1)
    inc a
    cp e
    ret nz
    dec de
    ret

    ; list length
    ; A=list -> X0
    ; modifies: X1, a, iy, A
llength:
    ld bc, 0
llen1:
    call fl_deref
    bit IMMEDIATE_BIT, l
    ret nz
    inc hl
    ld a, (hl)
    bit LIST_BIT, a
    ret z
    inc bc
    inc hl
    inc hl
    inc hl
    ld a, (hl)
    inc hl
    ld h, (hl)
    ld l, a
    jr llen1

fl_assign_3:
    ld hl, (PARAM1)
    ld iy, (PARAM0)
    call fl_assign
    ld hl, (PARAM2)
    ; fall through ...

    ; iy=var
confirm:
    ld l, lo(fl_nil)
    ld h, hi(fl_nil)
    jp fl_assign

fl_list_to_tuple_2:
    ld hl, (PARAM0)
    call llength
    ld a, c
    dec a
    ld hl, (PARAM0)
    call fl_mktuple0
    ld iy, (PARAM1)
    jp fl_unify_result

fl_halt_0:
    jp exit

fl_uncons_3: ;.
    ld hl, (PARAM0)
    call fl_check_cell
    push hl
    ld iy, (PARAM0)
    ld l, (iy + 2)
    ld h, (iy + 3)
    ld iy, (PARAM1)
    call fl_assign
    pop iy
    ld l, (iy + 4)
    ld h, (iy + 5)
    ld iy, (PARAM2)
    jp fl_assign

fl_string_to_list_3:
    push ix
    ld hl, (PARAM0)
    call fl_check_atom
    ld ix, (PARAM0)
    ld hl, (PARAM2)
    ld a, (ix)
    or a
    jr z, ls2l2
    ld d, 0
    ld e, a
    add ix, de
    ld b, a
ls2l1:
    push bc
    push hl
    ld h, 1 << (LOW_BIT - 1)
    ld a, (ix)
    add a, a
    rl h
    or 1
    ld l, a
    pop bc
    ld a, 1 << LIST_BIT
    call fl_cons
    pop bc
    dec ix
    djnz ls2l1    
ls2l2:
    ld iy, (PARAM1)
    pop ix
    jp fl_unify_result

    ; locate and check index into tuple
    ; A=tup, X0=index -> A
    ; modifies: iy, X1, a
index:
    push hl
    pop iy
    dec c
    jr z, lind2
    ld a, c
    and ~(ARITY_MASK >> 8)
    jr z, lind0
    ld de, EBADINDEX
    jp error
lind0:
    ld b, c
lind1:
    ld l, (iy + 4)
    ld h, (iy + 5)
    call fl_deref
    push hl
    pop iy
    djnz lind1
lind2:
    ld l, (iy + 2)
    ld h, (iy + 3)
    ret

fl_get_arg_3:
    ld hl, (PARAM0)
    call fl_getint
    ld b, h
    ld c, l
    ld hl, (PARAM1)
    call fl_check_tuple
    call index
    ld iy, (PARAM2)
    jp fl_unify_result

fl_put_arg_3:
    ld hl, (PARAM0)
    call fl_getint
    ld b, h
    ld c, l
    ld hl, (PARAM1)
    call fl_check_tuple
    call index
    push hl
    ld hl, (PARAM2)
    pop iy
    jp fl_assign

    ; chase stream tail and assign val 
    ; A=S, X0=val
    ; modifies: X1
chase_tail:
    ld (tmp1), hl
lct1:
    call fl_deref
    inc hl
    ld a, (hl)
    dec hl
    bit LIST_BIT, a
    jr z, lct2      ; loop until non-list (expected: var) 
    ld de, 4
    add hl, de
    ld e, (hl)
    inc hl
    ld d, (hl)
    ld hl, de
    jr lct1
lct2:
    push hl
    ld l, c
    ld h, b
    call fl_addref  ; addref new tail
    ld bc, (tmp1)
    call fl_unref    ; unref old S
    pop iy
    jp fl_assign   ; assign X to tail

    ; P0=[S|_], P1=X - chase tail in S and assign X
fl_merge_2:
    ld iy, (PARAM0)     ; get S
    ld l, (iy + 2)
    ld h, (iy + 3)
    ld bc, (PARAM1)
    call chase_tail
    ld bc, (PARAM1)
    ld iy, (PARAM0)         ; destructively change S to X
    ld (iy + 2), c
    ld (iy + 3), b
    ret

    ; P0=[S|Count], P1=Amount
fl_merge_refs_2:
    ld iy, (PARAM0)
    ld b, (iy + 4)
    dec b
    ld a, (PARAM1)
    add b
    ld (iy + 4), a
    ret

    ; P0=addr, P1=byte
fl_poke_2:
    ld hl, (PARAM0)
    call fl_check_integer
    call fl_getint
    push hl
    ld hl, (PARAM1)
    call fl_check_integer
    call fl_getint
    ld a, l
    pop hl
    ld (hl), a
    ret

    ; P0=addr, P1=result
fl_peek_2:
    ld hl, (PARAM0)
    call fl_check_integer
    call fl_getint
    ld a, (hl)
    ld h, 40h
    rla
    rl h
    or 1
    ld l, a
    ld iy, (PARAM1)
    jp fl_unify_result

; data area

ENOCELLS: db 'heap full$'
ENOGOALS: db 'goal buffer full$'
EQUEUEFULL: db 'process pool overflow$'
EBOUND: db 'assignment to non-variable$'
EFAIL: db 'failed$'
EUNIFY: db 'can not unify$'
EDIV0: db 'division by zero$'
EDEADLOCK: db 'deadlock$'
ELEAK: db 'memory leak$'
ENOINT: db 'expected integer$'
ENOSTRING: db 'expected string$'
ENOVAR: db 'expected variable$'
ENOTUPLE: db 'expected tuple$'
ENOLIST: db 'expected list$'
ENOCELL: db 'expected cell$'
EBADINDEX: db 'tuple index out of range$'
ESSFULL: db 'suspension-stack overflow$'
EARITY: db 'tuple length maximum exceeded$'

AP:     dw 0
NGOAL:  dw 0
PARAM0: dw 0
PARAM1: dw 0
PARAM2: dw 0
PARAM3: dw 0
ENV:    ds 2 * MAX_ENV

tmp1: dw 0
freelist: dw 0
freegoals: dw 0
goals:  dw 0
cells:  dw 0
sstackp: dw 0
suspended: dw 0
qstart: dw 0
qend:   dw 0
ticks:  dw 0
stackbase: dw 0
idle:   dw 0
epoch:  db 2
tailcalls:  db 0
timeslice:  db TIMESLICE

    align 2
    ds 1
fl_nil:
    db 2, '[]'
    align 2
    ds 1
fl_true:
    db 4, 'true'
    align 2
    ds 1
fl_false:
    db 5, 'false'
    ds 1

init:

goalbuf = (HEAP - (MAX_GOALS + 1) * 5 * 2) & ~255
sstack = goalbuf - 256
sstacktop = goalbuf
queuetop = sstack
queue = (queuetop - 2 * MAX_GOALS) & ~255

    end
